from __future__ import annotations

from typing import Any, TypedDict, cast
from typing_extensions import Literal, Optional

from deno_sandbox.timelines import Timeline

from .bridge import AsyncBridge
from .console import (
    AsyncConsoleClient,
    AsyncPaginatedList,
    PaginatedList,
)
from .utils import convert_to_snake_case


class RevisionWithoutTimelines(TypedDict):
    id: str
    """The unique identifier for the revision."""

    status: Literal["building", "ready", "error", "routed"]
    """The status of the revision."""

    created_at: str
    """The ISO 8601 timestamp when the revision was created."""

    updated_at: str
    """The ISO 8601 timestamp when the revision was last updated."""


class Revision(RevisionWithoutTimelines):
    timelines: list[Timeline]
    """The timelines associated with the revision."""


class AsyncRevisions:
    def __init__(self, client: AsyncConsoleClient):
        self._client = client

    async def get(self, app: str, id: str) -> Revision | None:
        """Get a revision by its ID for a specific app."""
        result = await self._client.get_or_none(f"/api/v2/apps/{app}/revisions/{id}")
        if result is None:
            return None
        raw_result = convert_to_snake_case(result)
        return cast(Revision, raw_result)

    async def list(
        self,
        app: str,
        *,
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> AsyncPaginatedList[RevisionWithoutTimelines]:
        """List revisions for a specific app.

        Args:
            app: The app ID or slug.
            cursor: The cursor for pagination.
            limit: Limit the number of items to return.
        """
        options: dict[str, Any] = {}
        if cursor is not None:
            options["cursor"] = cursor
        if limit is not None:
            options["limit"] = limit
        return await self._client.get_paginated(
            f"/api/v2/apps/{app}/revisions",
            cursor=None,
            params=options if options else None,
        )


class Revisions:
    def __init__(self, client: AsyncConsoleClient, bridge: AsyncBridge):
        self._client = client
        self._bridge = bridge
        self._async = AsyncRevisions(client)

    def get(self, app: str, id: str) -> Revision | None:
        """Get a revision by its ID for a specific app."""
        return self._bridge.run(self._async.get(app, id))

    def list(
        self,
        app: str,
        *,
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> PaginatedList[RevisionWithoutTimelines]:
        """List revisions for a specific app.

        Args:
            app: The app ID or slug.
            cursor: The cursor for pagination.
            limit: Limit the number of items to return.
        """
        paginated = self._bridge.run(self._async.list(app, cursor=cursor, limit=limit))
        return PaginatedList(self._bridge, paginated)



from __future__ import annotations

import base64
import builtins
from contextlib import asynccontextmanager, contextmanager
from datetime import datetime, timedelta, timezone
import json
from typing import (
    Any,
    AsyncIterator,
    BinaryIO,
    Iterable,
    Optional,
    TypedDict,
    Union,
    cast,
)
from typing_extensions import Literal, NotRequired, TypeAlias
import httpx

from .stream import Streamable, complete_stream, start_stream
from .utils import convert_to_snake_case
from .env import AsyncSandboxEnv, SandboxEnv
from .fs import AsyncSandboxFs, SandboxFs
from .process import (
    AbortSignal,
    AsyncChildProcess,
    AsyncDenoProcess,
    AsyncDenoRepl,
    ChildProcess,
    DenoProcess,
    DenoRepl,
    ProcessSpawnResult,
    RemoteProcessOptions,
)
from .bridge import AsyncBridge
from .console import (
    AsyncConsoleClient,
    AsyncPaginatedList,
    ExposeSSHResult,
    PaginatedList,
)
from .rpc import AsyncFetchResponse, AsyncRpcClient, FetchResponse
from .transport import (
    WebSocketTransport,
)
from .revisions import Revision


Mode: TypeAlias = Literal["connect", "create"]
StdIo: TypeAlias = Literal["piped", "null"]


class SandboxMeta(TypedDict):
    id: str
    """The unique identifier for the sandbox."""

    created_at: str
    """The ISO 8601 timestamp when the sandbox was created."""

    region: str
    """The region the sandbox is located in."""

    status: Literal["running", "stopped"]
    """The status of the sandbox."""

    stopped_at: NotRequired[str | None]
    """The ISO 8601 timestamp when the sandbox was stopped."""


class SecretConfig(TypedDict):
    """List of hostnames where this secret can be used. Must have at least one host."""

    hosts: builtins.list[str]
    value: str


class VolumeInfo(TypedDict):
    volume: str
    path: str


class AppConfigVolume(TypedDict):
    volume: str


class AppConfig(TypedDict):
    stop_at_ms: NotRequired[int | None]
    labels: NotRequired[dict[str, str] | None]
    memory_mb: NotRequired[int | None]
    volumes: NotRequired[builtins.list[VolumeInfo] | None]
    allow_net: NotRequired[builtins.list[str] | None]
    root: NotRequired[AppConfigVolume | None]
    secrets: NotRequired[dict[str, SecretConfig] | None]


class ExposeHTTPResult(TypedDict):
    domain: str


class BuildLog(TypedDict):
    """A build log entry from app deployment."""

    timestamp: str
    """The timestamp of the build log."""

    level: Literal["info", "error"]
    """The level of the build log."""

    message: str
    """The message of the build log."""


class AsyncSandboxApi:
    def __init__(
        self,
        client: AsyncConsoleClient,
    ):
        self._client = client

    @asynccontextmanager
    async def create(
        self,
        *,
        region: Optional[str] = None,
        env: Optional[dict[str, str]] = None,
        timeout: Optional[str] = None,
        memory_mb: Optional[int] = None,
        debug: Optional[bool] = None,
        labels: Optional[dict[str, str]] = None,
        root: Optional[str] = None,
        volumes: Optional[dict[str, str]] = None,
        allow_net: Optional[builtins.list[str]] = None,
        secrets: Optional[dict[str, SecretConfig]] = None,
        ssh: Optional[bool] = None,
        port: Optional[int] = None,
    ) -> AsyncIterator[AsyncSandbox]:
        """Creates a new sandbox instance.

        Args:
            region: The region where the sandbox should be created.
            env: Environment variables to start the sandbox with.
            timeout: The timeout of the sandbox. Defaults to "session". Other values like "30s" or "2m" are supported.
            memory_mb: The memory size in MiB of the sandbox. Defaults to 1280.
            debug: Enable debug logging for the sandbox connection.
            labels: Labels to set on the sandbox. Up to 5 labels can be specified.
            root: A volume or snapshot to use as the root filesystem of the sandbox.
            volumes: Volumes to mount on the sandbox. The key is the mount path, the value is the volume ID or slug.
            allow_net: List of hostnames/IP addresses the sandbox can make outbound requests to.
            secrets: Secret environment variables injected on the wire for HTTPS requests.
            ssh: Whether to expose SSH access to the sandbox.
            port: The port number to expose for HTTP access.
        """
        config_dict: dict[str, Any] = {
            "memory_mb": memory_mb if memory_mb is not None else 1280,
            "debug": debug if debug is not None else False,
        }

        if region is not None:
            config_dict["region"] = region
        if env is not None:
            config_dict["env"] = env
        if timeout is not None:
            config_dict["timeout"] = timeout
        if labels is not None:
            config_dict["labels"] = labels
        if root is not None:
            config_dict["root"] = {"volume": root}
        if volumes is not None:
            config_dict["volumes"] = volumes
        if allow_net is not None:
            config_dict["allow_net"] = allow_net
        if secrets is not None:
            config_dict["secrets"] = secrets
        if ssh is not None:
            config_dict["ssh"] = ssh
        if port is not None:
            config_dict["port"] = port

        app_config = cast(AppConfig, config_dict)

        json_config = json.dumps(app_config, separators=(",", ":")).encode("utf-8")

        url = self._client._options["sandbox_ws_url"].join("/api/v3/sandboxes/create")
        token = self._client._options["token"]

        transport = WebSocketTransport()
        ws = await transport.connect(
            url=url,
            headers={
                "Authorization": f"Bearer {token}",
                "x-deno-sandbox-config": base64.b64encode(json_config).decode("utf-8"),
            },
        )

        response = ws.response
        if response is None:
            raise Exception("No response received")

        if response.headers is None:
            raise Exception("No response headers received")

        sandbox_id = response.headers.get("x-deno-sandbox-id")
        if sandbox_id is None:
            raise Exception("Sandbox ID not found in response headers")

        sandbox = None
        try:
            rpc = AsyncRpcClient(transport)
            sandbox = AsyncSandbox(self._client, rpc, sandbox_id)
            yield sandbox
        finally:
            if sandbox is not None:
                await sandbox.close()

    @asynccontextmanager
    async def connect(
        self,
        sandbox_id: str,
    ) -> AsyncIterator[AsyncSandbox]:
        """Connects to an existing sandbox instance.

        Args:
            sandbox_id: The unique id of the sandbox to connect to.
        """
        url = self._client._options["sandbox_ws_url"].join(
            f"/api/v3/sandbox/{sandbox_id}/connect"
        )
        token = self._client._options["token"]
        transport = WebSocketTransport()
        await transport.connect(
            url=url,
            headers={
                "Authorization": f"Bearer {token}",
            },
        )

        sandbox = None
        try:
            rpc = AsyncRpcClient(transport)
            sandbox = AsyncSandbox(self._client, rpc, sandbox_id)
            yield sandbox
        finally:
            if sandbox is not None:
                await sandbox.close()

    async def list(
        self,
        *,
        labels: Optional[dict[str, str]] = None,
    ) -> AsyncPaginatedList[SandboxMeta]:
        """List sandboxes.

        Args:
            labels: Filter sandboxes by labels.
        """
        options: dict[str, Any] = {}
        if labels is not None:
            options["labels"] = labels
        return await self._client.get_paginated(
            path="/api/v3/sandboxes", cursor=None, params=options if options else None
        )


class SandboxApi:
    def __init__(self, client: AsyncConsoleClient, bridge: AsyncBridge):
        self._bridge = bridge
        self._client = client
        self._async = AsyncSandboxApi(client)

    @contextmanager
    def create(
        self,
        *,
        region: Optional[str] = None,
        env: Optional[dict[str, str]] = None,
        timeout: Optional[str] = None,
        memory_mb: Optional[int] = None,
        debug: Optional[bool] = None,
        labels: Optional[dict[str, str]] = None,
        root: Optional[str] = None,
        volumes: Optional[dict[str, str]] = None,
        allow_net: Optional[builtins.list[str]] = None,
        secrets: Optional[dict[str, SecretConfig]] = None,
        ssh: Optional[bool] = None,
        port: Optional[int] = None,
    ):
        """Creates a new sandbox instance.

        Args:
            region: The region where the sandbox should be created.
            env: Environment variables to start the sandbox with.
            timeout: The timeout of the sandbox. Defaults to "session". Other values like "30s" or "2m" are supported.
            memory_mb: The memory size in MiB of the sandbox. Defaults to 1280.
            debug: Enable debug logging for the sandbox connection.
            labels: Labels to set on the sandbox. Up to 5 labels can be specified.
            root: A volume or snapshot to use as the root filesystem of the sandbox.
            volumes: Volumes to mount on the sandbox. The key is the mount path, the value is the volume ID or slug.
            allow_net: List of hostnames/IP addresses the sandbox can make outbound requests to.
            secrets: Secret environment variables injected on the wire for HTTPS requests.
            ssh: Whether to expose SSH access to the sandbox.
            port: The port number to expose for HTTP access.
        """
        async_cm = self._async.create(
            region=region,
            env=env,
            timeout=timeout,
            memory_mb=memory_mb,
            debug=debug,
            labels=labels,
            root=root,
            volumes=volumes,
            allow_net=allow_net,
            secrets=secrets,
            ssh=ssh,
            port=port,
        )
        async_handle = self._bridge.run(async_cm.__aenter__())

        try:
            yield Sandbox(self._client, self._bridge, async_handle._rpc, async_handle)
        except Exception:
            import sys

            self._bridge.run(async_cm.__aexit__(*sys.exc_info()))
            raise
        finally:
            self._bridge.run(async_cm.__aexit__(None, None, None))

    @contextmanager
    def connect(
        self,
        sandbox_id: str,
    ):
        """Connects to an existing sandbox instance.

        Args:
            sandbox_id: The unique id of the sandbox to connect to.
            region: If the sandbox was created in a non-default region, the region where the sandbox is running.
            debug: Enable debug logging for the sandbox connection.
            ssh: Whether to expose SSH access to the sandbox.
        """
        async_cm = self._async.connect(sandbox_id)
        async_handle = self._bridge.run(async_cm.__aenter__())

        try:
            yield Sandbox(self._client, self._bridge, async_handle._rpc, async_handle)
        except Exception:
            import sys

            self._bridge.run(async_cm.__aexit__(*sys.exc_info()))
            raise
        finally:
            self._bridge.run(async_cm.__aexit__(None, None, None))

    def list(
        self,
        *,
        labels: Optional[dict[str, str]] = None,
    ) -> PaginatedList[SandboxMeta]:
        """List sandboxes.

        Args:
            labels: Filter sandboxes by labels.
        """
        paginated = self._bridge.run(self._async.list(labels=labels))
        return PaginatedList(self._bridge, paginated)


async def _parse_sse_stream(stream: AsyncIterator[bytes]) -> AsyncIterator[str]:
    """Parse Server-Sent Events from a byte stream."""
    buffer = b""
    async for chunk in stream:
        buffer += chunk
        while b"\n" in buffer:
            line, buffer = buffer.split(b"\n", 1)
            line = line.rstrip(b"\r")
            if line.startswith(b"data: "):
                data = line[6:].decode("utf-8")
                yield data


class AsyncBuild:
    """The result of a deno.deploy() operation."""

    def __init__(
        self,
        revision_id: str,
        app: str,
        client: AsyncConsoleClient,
    ):
        self.id = revision_id
        """The ID of the build."""
        self._app = app
        self._client = client

    async def wait(self) -> Revision:
        """A coroutine that resolves when the build is complete, returning the revision."""

        url = self._client._options["console_url"].join(
            f"/api/v2/apps/{self._app}/revisions/{self.id}/status"
        )
        headers = {
            "Authorization": f"Bearer {self._client._options['token']}",
        }

        async with httpx.AsyncClient() as http_client:
            response = await http_client.get(str(url), headers=headers)
            response.raise_for_status()
            revision_data = response.json()

            # Fetch timelines
            timelines_url = self._client._options["console_url"].join(
                f"/api/v2/apps/{self._app}/revisions/{self.id}/timelines"
            )
            timelines_response = await http_client.get(
                str(timelines_url), headers=headers
            )
            timelines_response.raise_for_status()
            timelines_data = timelines_response.json()

            result = convert_to_snake_case(revision_data)
            result["timelines"] = convert_to_snake_case(timelines_data)
            return cast(Revision, result)

    async def logs(self) -> AsyncIterator[BuildLog]:
        """An async iterator of build logs."""
        url = self._client._options["console_url"].join(
            f"/api/v2/apps/{self._app}/revisions/{self.id}/logs"
        )
        headers = {
            "Authorization": f"Bearer {self._client._options['token']}",
        }

        async with httpx.AsyncClient() as http_client:
            async with http_client.stream("GET", str(url), headers=headers) as response:
                response.raise_for_status()
                async for data in _parse_sse_stream(response.aiter_bytes()):
                    try:
                        yield cast(BuildLog, json.loads(data))
                    except json.JSONDecodeError:
                        # Skip malformed log entries
                        continue


class Build:
    """The result of a deno.deploy() operation (sync version)."""

    def __init__(
        self,
        revision_id: str,
        app: str,
        client: AsyncConsoleClient,
        bridge: AsyncBridge,
    ):
        self.id = revision_id
        """The ID of the build."""
        self._async_build = AsyncBuild(revision_id, app, client)
        self._bridge = bridge

    def wait(self) -> Revision:
        """Returns the revision when the build is complete."""
        return self._bridge.run(self._async_build.wait())

    def logs(self) -> builtins.list[BuildLog]:
        """Returns a list of build logs."""

        async def _collect_logs():
            logs = []
            async for log in self._async_build.logs():
                logs.append(log)
            return logs

        return self._bridge.run(_collect_logs())


class AsyncSandboxDeno:
    def __init__(
        self,
        rpc: AsyncRpcClient,
        processes: builtins.list[AsyncChildProcess],
        client: AsyncConsoleClient,
        sandbox_id: str,
    ):
        self._rpc = rpc
        self._processes = processes
        self._client = client
        self._sandbox_id = sandbox_id

    async def run(
        self,
        *,
        args: Optional[builtins.list[str]] = None,
        cwd: Optional[str] = None,
        clear_env: Optional[bool] = None,
        env: Optional[dict[str, str]] = None,
        signal: Optional[AbortSignal] = None,
        stdin: Optional[Literal["piped", "null"]] = None,
        stdout: Optional[Literal["piped", "null", "inherit"]] = None,
        stderr: Optional[Literal["piped", "null", "inherit"]] = None,
        script_args: Optional[list[str]] = None,
        entrypoint: Optional[str] = None,
        code: Optional[str] = None,
        extension: Optional[
            Literal["js", "cjs", "mjs", "ts", "cts", "mts", "jsx", "tsx"]
        ] = None,
        stdin_data: Optional[Streamable] = None,
    ) -> AsyncDenoProcess:
        """Create a new Deno process from the specified entrypoint file or code.

        Args:
            args: Arguments to pass to the process.
            cwd: The working directory of the process.
            clear_env: Clear environment variables from parent process.
            env: Environment variables to pass to the subprocess.
            signal: An abort signal to cancel the process.
            stdin: How stdin of the spawned process should be handled.
            stdout: How stdout of the spawned process should be handled.
            stderr: How stderr of the spawned process should be handled.
            script_args: Arguments to pass to the Deno runtime, available as Deno.args.
            entrypoint: A module to read from disk and execute as the entrypoint.
            code: Deno code to execute as the entrypoint.
            extension: File extension to use when executing code. Default is 'ts'.
            stdin_data: Data to write to stdin of the process.
        """
        params: dict[str, Any] = {
            "stdout": stdout if stdout is not None else "inherit",
            "stderr": stderr if stderr is not None else "inherit",
        }

        if args is not None:
            params["args"] = args
        if cwd is not None:
            params["cwd"] = cwd
        if clear_env is not None:
            params["clear_env"] = clear_env
        if env is not None:
            params["env"] = env
        if signal is not None:
            params["signal"] = signal
        if stdin is not None:
            params["stdin"] = stdin
        if script_args is not None:
            params["script_args"] = script_args
        if entrypoint is not None:
            params["entrypoint"] = entrypoint
        if code is not None:
            params["code"] = code
        if extension is not None:
            params["extension"] = extension

        if "code" in params and "extension" not in params:
            params["extension"] = "ts"

        # If stdin data is provided, start stream first (but don't send data yet)
        stdin_writer = None
        if stdin_data is not None:
            params["stdin"] = "piped"
            stdin_stream_id, stdin_writer = await start_stream(self._rpc)
            params["stdinStreamId"] = stdin_stream_id

        opts = RemoteProcessOptions(
            stdout_inherit=params["stdout"] == "inherit",
            stderr_inherit=params["stderr"] == "inherit",
        )

        if params["stdout"] == "inherit":
            params["stdout"] = "piped"
        if params["stderr"] == "inherit":
            params["stderr"] = "piped"

        result = await self._rpc.call("spawnDeno", params)

        # Now that process is spawned, complete the stdin stream
        if stdin_writer is not None and stdin_data is not None:
            await complete_stream(stdin_writer, stdin_data)

        process = await AsyncDenoProcess.create(
            result, self._rpc, opts, self._processes
        )
        self._processes.append(process)
        return process

    async def eval(self, code: str) -> Any:
        repl = await self.repl()
        result = await repl.eval(code)
        await repl.close()
        return result

    async def repl(
        self,
        *,
        args: Optional[builtins.list[str]] = None,
        cwd: Optional[str] = None,
        clear_env: Optional[bool] = None,
        env: Optional[dict[str, str]] = None,
        signal: Optional[AbortSignal] = None,
        stdin: Optional[Literal["piped", "null"]] = None,
        stdout: Optional[Literal["piped", "null", "inherit"]] = None,
        stderr: Optional[Literal["piped", "null", "inherit"]] = None,
        script_args: Optional[builtins.list[str]] = None,
    ) -> AsyncDenoRepl:
        """Create a new Deno REPL process.

        Args:
            args: Arguments to pass to the process.
            cwd: The working directory of the process.
            clear_env: Clear environment variables from parent process.
            env: Environment variables to pass to the subprocess.
            signal: An abort signal to cancel the process.
            stdin: How stdin of the spawned process should be handled.
            stdout: How stdout of the spawned process should be handled.
            stderr: How stderr of the spawned process should be handled.
            script_args: Arguments to pass to the Deno runtime, available as Deno.args.
        """
        params: dict[str, Any] = {"stdout": "piped", "stderr": "piped"}

        opts = RemoteProcessOptions(stdout_inherit=True, stderr_inherit=True)

        if args is not None:
            params["args"] = args
        if cwd is not None:
            params["cwd"] = cwd
        if clear_env is not None:
            params["clearEnv"] = clear_env
        if env is not None:
            params["env"] = env
        if signal is not None:
            params["signal"] = signal
        if stdin is not None:
            params["stdin"] = stdin
        if stdout is not None:
            params["stdout"] = stdout
            if stdout != "inherit":
                opts["stdout_inherit"] = False
        if stderr is not None:
            params["stderr"] = stderr
            if stderr != "inherit":
                opts["stderr_inherit"] = False
        if script_args is not None:
            params["scriptArgs"] = script_args

        result: ProcessSpawnResult = await self._rpc.call("spawnDenoRepl", params)

        process = await AsyncDenoRepl.create(result, self._rpc, opts, self._processes)
        self._processes.append(process)
        return process

    async def deploy(
        self,
        app: str,
        *,
        entrypoint: Optional[str] = None,
        args: Optional[builtins.list[str]] = None,
        path: Optional[str] = None,
        production: Optional[bool] = None,
        preview: Optional[bool] = None,
    ) -> AsyncBuild:
        """Deploy the contents of the sandbox to the specified app in Deno Deploy platform.

        Args:
            app: The app ID or slug to deploy to.
            entrypoint: The entrypoint file path relative to the path option. Defaults to 'main.ts'.
            args: Arguments to pass to the entrypoint script.
            path: The path to the directory to deploy. If relative, it is relative to /app. Defaults to '/app'.
            production: Whether to deploy in production mode. Defaults to True.
            preview: Whether to deploy a preview deployment. Defaults to False.

        Returns:
            An AsyncBuild object with the revision ID and methods to check status and logs.

        Example:
            ```python
            from deno_sandbox import AsyncDenoDeploy

            async with AsyncDenoDeploy() as client:
                async with client.sandbox.create() as sandbox:
                    await sandbox.fs.write_text_file(
                        "main.ts",
                        'Deno.serve(() => new Response("Hi from sandbox.deploy()"))',
                    )
                    build = await sandbox.deno.deploy("my-deno-app", entrypoint="main.ts")
                    print(f"Deployed revision ID: {build.id}")
                    revision = await build.done
                    print(f"Revision status: {revision['status']}")
            ```
        """
        url = self._client._options["console_url"].join(f"/api/v2/apps/{app}/deploy")
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self._client._options['token']}",
        }

        # Build request body
        body: dict[str, Any] = {
            "entrypoint": entrypoint if entrypoint is not None else "main.ts",
            "sandboxId": self._sandbox_id,
            "path": path if path is not None else "/app",
        }

        if args is not None:
            body["args"] = args
        if production is not None:
            body["production"] = production
        if preview is not None:
            body["preview"] = preview

        # Make the deploy request
        async with httpx.AsyncClient() as http_client:
            response = await http_client.post(
                str(url), headers=headers, json=body, timeout=30.0
            )
            response.raise_for_status()
            result = response.json()
            revision_id = result["revisionId"]

        return AsyncBuild(revision_id, app, self._client)


class SandboxDeno:
    def __init__(
        self,
        rpc: AsyncRpcClient,
        bridge: AsyncBridge,
        processes: builtins.list[AsyncChildProcess],
        client: AsyncConsoleClient,
        sandbox_id: str,
    ):
        self._rpc = rpc
        self._bridge = bridge
        self._client = client

        self._async = AsyncSandboxDeno(rpc, processes, client, sandbox_id)

    def run(
        self,
        *,
        args: Optional[builtins.list[str]] = None,
        cwd: Optional[str] = None,
        clear_env: Optional[bool] = None,
        env: Optional[dict[str, str]] = None,
        signal: Optional[AbortSignal] = None,
        stdin: Optional[Literal["piped", "null"]] = None,
        stdout: Optional[Literal["piped", "null", "inherit"]] = None,
        stderr: Optional[Literal["piped", "null", "inherit"]] = None,
        script_args: Optional[builtins.list[str]] = None,
        entrypoint: Optional[str] = None,
        code: Optional[str] = None,
        extension: Optional[
            Literal["js", "cjs", "mjs", "ts", "cts", "mts", "jsx", "tsx"]
        ] = None,
        stdin_data: Optional[Union[Iterable[bytes], BinaryIO]] = None,
    ) -> DenoProcess:
        """Create a new Deno process from the specified entrypoint file or code.

        Args:
            args: Arguments to pass to the process.
            cwd: The working directory of the process.
            clear_env: Clear environment variables from parent process.
            env: Environment variables to pass to the subprocess.
            signal: An abort signal to cancel the process.
            stdin: How stdin of the spawned process should be handled.
            stdout: How stdout of the spawned process should be handled.
            stderr: How stderr of the spawned process should be handled.
            script_args: Arguments to pass to the Deno runtime, available as Deno.args.
            entrypoint: A module to read from disk and execute as the entrypoint.
            code: Deno code to execute as the entrypoint.
            extension: File extension to use when executing code. Default is 'ts'.
            stdin_data: Data to write to stdin of the process.
        """
        async_deno = self._bridge.run(
            self._async.run(
                args=args,
                cwd=cwd,
                clear_env=clear_env,
                env=env,
                signal=signal,
                stdin=stdin,
                stdout=stdout,
                stderr=stderr,
                script_args=script_args,
                entrypoint=entrypoint,
                code=code,
                extension=extension,
                stdin_data=stdin_data,
            )
        )
        return DenoProcess(self._rpc, self._bridge, async_deno)

    def eval(self, code: str) -> Any:
        return self._bridge.run(self._async.eval(code))

    def repl(
        self,
        *,
        args: Optional[builtins.list[str]] = None,
        cwd: Optional[str] = None,
        clear_env: Optional[bool] = None,
        env: Optional[dict[str, str]] = None,
        signal: Optional[AbortSignal] = None,
        stdin: Optional[Literal["piped", "null"]] = None,
        stdout: Optional[Literal["piped", "null", "inherit"]] = None,
        stderr: Optional[Literal["piped", "null", "inherit"]] = None,
        script_args: Optional[builtins.list[str]] = None,
    ) -> DenoRepl:
        """Create a new Deno REPL process.

        Args:
            args: Arguments to pass to the process.
            cwd: The working directory of the process.
            clear_env: Clear environment variables from parent process.
            env: Environment variables to pass to the subprocess.
            signal: An abort signal to cancel the process.
            stdin: How stdin of the spawned process should be handled.
            stdout: How stdout of the spawned process should be handled.
            stderr: How stderr of the spawned process should be handled.
            script_args: Arguments to pass to the Deno runtime, available as Deno.args.
        """
        async_repl = self._bridge.run(
            self._async.repl(
                args=args,
                cwd=cwd,
                clear_env=clear_env,
                env=env,
                signal=signal,
                stdin=stdin,
                stdout=stdout,
                stderr=stderr,
                script_args=script_args,
            )
        )
        return DenoRepl(self._rpc, self._bridge, async_repl)

    def deploy(
        self,
        app: str,
        *,
        entrypoint: Optional[str] = None,
        args: Optional[builtins.list[str]] = None,
        path: Optional[str] = None,
        production: Optional[bool] = None,
        preview: Optional[bool] = None,
    ) -> Build:
        """Deploy the contents of the sandbox to the specified app in Deno Deploy platform.

        Args:
            app: The app ID or slug to deploy to.
            entrypoint: The entrypoint file path relative to the path option. Defaults to 'main.ts'.
            args: Arguments to pass to the entrypoint script.
            path: The path to the directory to deploy. If relative, it is relative to /app. Defaults to '/app'.
            production: Whether to deploy in production mode. Defaults to True.
            preview: Whether to deploy a preview deployment. Defaults to False.

        Returns:
            A Build object with the revision ID and methods to check status and logs.

        Example:
            ```python
            from deno_sandbox import DenoDeploy

            client = DenoDeploy()
            with client.sandbox.create() as sandbox:
                sandbox.fs.write_text_file(
                    "main.ts",
                    'Deno.serve(() => new Response("Hi from sandbox.deploy()"))',
                )
                build = sandbox.deno.deploy("my-deno-app", entrypoint="main.ts")
                print(f"Deployed revision ID: {build.id}")
                revision = build.done
                print(f"Revision status: {revision['status']}")
            ```
        """
        async_build = self._bridge.run(
            self._async.deploy(
                app,
                entrypoint=entrypoint,
                args=args,
                path=path,
                production=production,
                preview=preview,
            )
        )
        return Build(async_build.id, app, self._client, self._bridge)


class AsyncSandbox:
    def __init__(
        self, client: AsyncConsoleClient, rpc: AsyncRpcClient, sandbox_id: str
    ):
        self._client = client
        self._rpc = rpc
        self._processes: builtins.list[AsyncChildProcess] = []

        self.url: str | None = None
        self.ssh: None = None
        self.id = sandbox_id
        self.fs = AsyncSandboxFs(rpc)
        self.deno = AsyncSandboxDeno(rpc, self._processes, client, sandbox_id)
        self.env = AsyncSandboxEnv(rpc)

    @property
    def closed(self) -> bool:
        return self._rpc._transport.closed

    async def spawn(
        self,
        command: str,
        *,
        args: Optional[list[str]] = None,
        cwd: Optional[str] = None,
        clear_env: Optional[bool] = None,
        env: Optional[dict[str, str]] = None,
        signal: Optional[AbortSignal] = None,
        stdin: Optional[Literal["piped", "null"]] = None,
        stdout: Optional[Literal["piped", "null", "inherit"]] = None,
        stderr: Optional[Literal["piped", "null", "inherit"]] = None,
        stdin_data: Optional[Streamable] = None,
    ) -> AsyncChildProcess:
        """Spawn a new child process.

        Args:
            command: The command to execute.
            args: Arguments to pass to the process.
            cwd: The working directory of the process.
            clear_env: Clear environment variables from parent process.
            env: Environment variables to pass to the subprocess.
            signal: An abort signal to cancel the process.
            stdin: How stdin of the spawned process should be handled.
            stdout: How stdout of the spawned process should be handled.
            stderr: How stderr of the spawned process should be handled.
            stdin_data: Data to write to stdin of the process.
        """
        params: dict[str, Any] = {
            "command": command,
            "stdout": stdout if stdout is not None else "inherit",
            "stderr": stderr if stderr is not None else "inherit",
        }

        if args is not None:
            params["args"] = args
        if cwd is not None:
            params["cwd"] = cwd
        if clear_env is not None:
            params["clear_env"] = clear_env
        if env is not None:
            params["env"] = env
        if signal is not None:
            params["signal"] = signal
        if stdin is not None:
            params["stdin"] = stdin

        # If stdin data is provided, start stream first (but don't send data yet)
        stdin_writer = None
        if stdin_data is not None:
            params["stdin"] = "piped"
            stdin_stream_id, stdin_writer = await start_stream(self._rpc)
            params["stdinStreamId"] = stdin_stream_id

        opts = RemoteProcessOptions(
            stdout_inherit=params["stdout"] == "inherit",
            stderr_inherit=params["stderr"] == "inherit",
        )

        if params["stdout"] == "inherit":
            params["stdout"] = "piped"
        if params["stderr"] == "inherit":
            params["stderr"] = "piped"

        result: ProcessSpawnResult = await self._rpc.call("spawn", params)

        # Now that process is spawned, complete the stdin stream
        if stdin_writer is not None and stdin_data is not None:
            await complete_stream(stdin_writer, stdin_data)

        process = await AsyncChildProcess.create(
            result, self._rpc, opts, self._processes
        )
        self._processes.append(process)
        return process

    async def fetch(
        self,
        url: str,
        *,
        method: Optional[str] = "GET",
        headers: Optional[dict[str, str]] = None,
        redirect: Optional[Literal["follow", "manual"]] = None,
    ) -> AsyncFetchResponse:
        return await self._rpc.fetch(url, method, headers, redirect)

    async def close(self) -> None:
        # Kill all tracked processes
        for process in self._processes:
            await process.kill()
        self._processes.clear()
        await self._rpc.close()

    async def kill(self) -> None:
        await self._client.delete(f"/api/v3/sandboxes/{self.id}")

    async def extend_timeout(self, additional_s: int) -> datetime:
        """Request to extend the timeout of the sandbox by the specified duration.
        You can at max extend timeout of a sandbox by 30 minutes at once.

        Please note the extension is not guranteed to be the same as requested time.
        You should rely on the returned Date value to know the exact extension time.
        """

        now = datetime.now(timezone.utc)
        future_time = now + timedelta(seconds=additional_s)
        stop_at_ms = int(future_time.timestamp() * 1000)

        url = self._client._options["sandbox_url"].join(f"/api/v3/sandbox/{self.id}")
        result = await self._client._request("PATCH", url, {"stop_at_ms": stop_at_ms})
        data = result.json()

        return datetime.fromtimestamp(data["stop_at_ms"] / 1000, tz=timezone.utc)

    async def expose_http(
        self, *, port: Optional[int] = None, pid: Optional[int] = None
    ) -> str:
        """Publicly expose a HTTP service via a publicly routeable URL.

        NOTE: when you call this API, the target HTTP service will be PUBLICLY
        EXPOSED WITHOUT AUTHENTICATION. Anyone with knowledge of the public domain
        will be able to send arbitrary requests to the exposed service.

        An exposed service can either be a service listening on an arbitrary HTTP
        port, or a JavaScript runtime that can handle HTTP requests.
        """

        if port is not None and pid is not None:
            raise ValueError("Only one of port or pid can be specified")

        params: dict[str, Any] = {}
        if port is not None:
            params["port"] = port
        if pid is not None:
            params["pid"] = pid

        url = self._client._options["sandbox_url"].join(
            f"/api/v3/sandbox/{self.id}/expose/http"
        )
        result = await self._client._request("POST", url, params)
        data = cast(ExposeHTTPResult, result.json())
        domain = data["domain"]

        params["domain"] = domain
        await self._rpc.call("exposeHttp", params)

        return f"https://{domain}"

    async def expose_ssh(self) -> ExposeSSHResult:
        """Expose an isolate over SSH, allowing access to the isolate's shell.

        NOTE: The SSH connection is authenticated through the 'username' field. This field is populated
        with a randomly generated, unique identifier. Anyone with knowledge of the 'username' can
        connect to the isolate's shell without further authentication.
        """

        url = self._client._options["sandbox_url"].join(
            f"/api/v3/sandbox/{self.id}/expose/ssh"
        )
        response = await self._client._request("POST", url, {})

        return cast(ExposeSSHResult, response.json())

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()


class Sandbox:
    def __init__(
        self,
        client: AsyncConsoleClient,
        bridge: AsyncBridge,
        rpc: AsyncRpcClient,
        async_sandbox: AsyncSandbox,
    ):
        self._client = client
        self._bridge = bridge
        self._rpc = rpc
        self._async = async_sandbox

        self.url: str | None = None
        self.ssh: None = None
        self.id = async_sandbox.id
        self.fs = SandboxFs(rpc, bridge)
        self.deno = SandboxDeno(
            rpc, bridge, self._async._processes, client, async_sandbox.id
        )
        self.env = SandboxEnv(rpc, bridge)

    @property
    def closed(self) -> bool:
        return self._rpc._transport.closed

    def spawn(
        self,
        command: str,
        *,
        args: Optional[builtins.list[str]] = None,
        cwd: Optional[str] = None,
        clear_env: Optional[bool] = None,
        env: Optional[dict[str, str]] = None,
        signal: Optional[AbortSignal] = None,
        stdin: Optional[Literal["piped", "null"]] = None,
        stdout: Optional[Literal["piped", "null", "inherit"]] = None,
        stderr: Optional[Literal["piped", "null", "inherit"]] = None,
        stdin_data: Optional[Union[Iterable[bytes], BinaryIO]] = None,
    ) -> ChildProcess:
        """Spawn a new child process.

        Args:
            command: The command to execute.
            args: Arguments to pass to the process.
            cwd: The working directory of the process.
            clear_env: Clear environment variables from parent process.
            env: Environment variables to pass to the subprocess.
            signal: An abort signal to cancel the process.
            stdin: How stdin of the spawned process should be handled.
            stdout: How stdout of the spawned process should be handled.
            stderr: How stderr of the spawned process should be handled.
            stdin_data: Data to write to stdin of the process.
        """
        async_child = self._bridge.run(
            self._async.spawn(
                command,
                args=args,
                cwd=cwd,
                clear_env=clear_env,
                env=env,
                signal=signal,
                stdin=stdin,
                stdout=stdout,
                stderr=stderr,
                stdin_data=stdin_data,
            )
        )
        return ChildProcess(self._rpc, self._bridge, async_child)

    def fetch(
        self,
        url: str,
        *,
        method: Optional[str] = "GET",
        headers: Optional[dict[str, str]] = None,
        redirect: Optional[Literal["follow", "manual"]] = None,
    ) -> FetchResponse:
        async_response = self._bridge.run(
            self._rpc.fetch(url, method, headers, redirect, None)
        )
        return FetchResponse(async_response)

    def close(self) -> None:
        self._bridge.run(self._async.close())

    def kill(self) -> None:
        self._bridge.run(self._async.kill())

    def extend_timeout(self, additional_s: int) -> datetime:
        """Request to extend the timeout of the sandbox by the specified duration.
        You can at max extend timeout of a sandbox by 30 minutes at once.

        Please note the extension is not guranteed to be the same as requested time.
        You should rely on the returned Date value to know the exact extension time.
        """
        return self._bridge.run(self._async.extend_timeout(additional_s))

    def expose_http(
        self, *, port: Optional[int] = None, pid: Optional[int] = None
    ) -> str:
        """Publicly expose a HTTP service via a publicly routeable URL.

        NOTE: when you call this API, the target HTTP service will be PUBLICLY
        EXPOSED WITHOUT AUTHENTICATION. Anyone with knowledge of the public domain
        will be able to send arbitrary requests to the exposed service.

        An exposed service can either be a service listening on an arbitrary HTTP
        port, or a JavaScript runtime that can handle HTTP requests.
        """
        return self._bridge.run(self._async.expose_http(port=port, pid=pid))

    def expose_ssh(self) -> ExposeSSHResult:
        """Expose an isolate over SSH, allowing access to the isolate's shell.

        NOTE: The SSH connection is authenticated through the 'username' field. This field is populated
        with a randomly generated, unique identifier. Anyone with knowledge of the 'username' can
        connect to the isolate's shell without further authentication.
        """
        return self._bridge.run(self._async.expose_ssh())

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._bridge.run(self._async.__aexit__(exc_type, exc_val, exc_tb))



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .rpc import AsyncRpcClient
    from .bridge import AsyncBridge


class AsyncSandboxEnv:
    def __init__(self, rpc: AsyncRpcClient):
        self._rpc = rpc

    async def get(self, key: str) -> str:
        """Get the value of an environment variable."""

        params = {"key": key}
        result = await self._rpc.call("envGet", params)

        return result

    async def set(self, key: str, value: str) -> None:
        """Set the value of an environment variable."""

        params = {"key": key, "value": value}
        await self._rpc.call("envSet", params)

    async def as_dict(self) -> dict[str, str]:
        """Get all environment variables."""

        params = {}
        result = await self._rpc.call("envToObject", params)

        return result

    async def delete(self, key: str) -> None:
        """Delete an environment variable."""

        params = {"key": key}
        await self._rpc.call("envDelete", params)


class SandboxEnv:
    def __init__(self, rpc: AsyncRpcClient, bridge: AsyncBridge):
        self._rpc = rpc
        self._bridge = bridge
        self._async = AsyncSandboxEnv(rpc)

    def get(self, key: str) -> str:
        """Get the value of an environment variable."""
        return self._bridge.run(self._async.get(key))

    def set(self, key: str, value: str) -> None:
        """Set the value of an environment variable."""
        self._bridge.run(self._async.set(key, value))

    def as_dict(self) -> dict[str, str]:
        """Get all environment variables."""
        return self._bridge.run(self._async.as_dict())

    def delete(self, key: str) -> None:
        """Delete an environment variable."""
        self._bridge.run(self._async.delete(key))



from __future__ import annotations

import os

from typing import Optional, TypedDict
from typing_extensions import NotRequired

from httpx import URL

from .errors import MissingApiToken


class Options(TypedDict):
    token: NotRequired[str | None]
    regions: NotRequired[list[str] | None]


class InternalOptions(TypedDict):
    sandbox_ws_url: URL
    sandbox_url: URL
    console_url: URL
    token: str
    regions: list[str]


def get_internal_options(options: Optional[Options] = None) -> InternalOptions:
    sandbox_url = URL(
        os.environ.get("DENO_SANDBOX_ENDPOINT", "https://ams.sandbox-api.deno.net")
    )

    token = (
        options is not None
        and options.get("token")
        or os.environ.get("DENO_DEPLOY_TOKEN", None)
    )

    if token is None:
        raise MissingApiToken({"message": "DENO_DEPLOY_TOKEN is not set"})

    scheme = sandbox_url.scheme.replace("http", "ws")
    sandbox_ws_url = URL(f"{scheme}://{sandbox_url.netloc.decode()}")

    console_url = URL(
        os.environ.get("DENO_DEPLOY_ENDPOINT", "https://console.deno.com")
    )

    regions = (
        options is not None
        and options.get("regions")
        or os.environ.get("DENO_AVAILABLE_REGIONS", "ams1,ord").split(",")
    )

    return InternalOptions(
        console_url=console_url,
        sandbox_ws_url=sandbox_ws_url,
        sandbox_url=sandbox_url,
        token=token,
        regions=regions,
    )



from __future__ import annotations

import asyncio
import base64
import json
from typing import Any, Dict, Literal, Mapping, Optional, TypedDict, cast
from typing_extensions import NotRequired
from websockets import ConnectionClosed

from .errors import (
    HTTPStatusError,
    ProcessAlreadyExited,
    RpcValidationError,
    UnknownRpcMethod,
    ZodErrorRaw,
)
from .transport import WebSocketTransport
from .utils import (
    convert_to_camel_case,
    convert_to_snake_case,
    to_snake_case,
)


class RpcRequest(TypedDict):
    id: int
    method: str
    params: dict[str, Any]
    jsonrpc: str


class RpcResult(TypedDict):
    ok: NotRequired[Any]
    error: NotRequired[Any]


class RpcResponse(TypedDict):
    id: int
    jsonrpc: str
    result: NotRequired[RpcResult | None]
    error: NotRequired[dict[str, Any] | None]


class AsyncRpcClient:
    def __init__(self, transport: WebSocketTransport):
        self._transport = transport
        self._id = 0
        self._pending_requests: Dict[int, asyncio.Future[Any]] = {}
        self._listen_task: asyncio.Task[Any] | None = None
        self._pending_processes: Dict[int, asyncio.StreamReader] = {}
        self.__loop: asyncio.AbstractEventLoop | None = None
        self._signal_id = 0
        self._stream_id = 0

    @property
    def _loop(self) -> asyncio.AbstractEventLoop:
        if self.__loop is None:
            self.__loop = asyncio.get_running_loop()
        return self.__loop

    async def close(self):
        await self._transport.close()

    async def send_notification(self, method: str, params: dict[str, Any]) -> None:
        """Send a notification (no response expected)."""
        payload = {"method": method, "params": params, "jsonrpc": "2.0"}
        await self._transport.send(json.dumps(payload))

    def next_stream_id(self) -> int:
        """Get next stream ID."""
        self._stream_id += 1
        return self._stream_id

    async def call(self, method: str, params: Mapping[str, Any]) -> Any:
        if self._listen_task is None or self._listen_task.done():
            self._listen_task = self._loop.create_task(self._listener())

        req_id = self._id + 1
        self._id = req_id

        camel_params = convert_to_camel_case(params)
        payload = RpcRequest(
            method=method, params=camel_params, id=req_id, jsonrpc="2.0"
        )

        future = self._loop.create_future()
        self._pending_requests[req_id] = future

        await self._transport.send(json.dumps(payload))

        raw_response = await future
        response = cast(RpcResponse, raw_response)

        maybeError = response.get("error")
        if maybeError is not None:
            if maybeError.get("message") == "Method not found":
                raise UnknownRpcMethod("RPC method not found")

            if maybeError.get("data") is not None:
                data = maybeError["data"]
                if data.get("constructor_name") == "ZodError":
                    # For some reason ZodError data is serialized as
                    # json inside json ¯\_(ツ)_/¯
                    zod_errors: list[ZodErrorRaw] = []
                    for e in json.loads(data["message"]):
                        value = cast(ZodErrorRaw, e)

                        value["path"] = [to_snake_case(p) for p in value["path"]]
                        zod_errors.append(value)

                    raise RpcValidationError(zod_errors)

            raise Exception(maybeError)

        maybeResult = response.get("result")
        if maybeResult is not None:
            err = maybeResult.get("error")
            if err is not None:
                if (
                    "constructor_name" in err
                    and err.get("constructor_name") == "TypeError"
                    and "code" in err
                    and err.get("code") == "ENOENT"
                ):
                    raise ProcessAlreadyExited("Process has already exited")

                raise Exception(f"Application Error: {err}")

        return maybeResult["ok"] if maybeResult else None

    async def _listener(self) -> None:
        try:
            async for raw in self._transport:
                data = json.loads(raw)
                req_id = data.get("id")

                if req_id is not None and req_id in self._pending_requests:
                    future = self._pending_requests.pop(req_id)
                    if not future.done():
                        converted_data = convert_to_snake_case(data)
                        future.set_result(converted_data)

                elif "method" in data:
                    method = data["method"]
                    params = data.get("params", {})

                    if method == "$sandbox.stream.enqueue":
                        stream_id = params.get("streamId")
                        chunk = base64.b64decode(params.get("data", ""))
                        stream = self._pending_processes.get(stream_id)
                        if stream:
                            stream.feed_data(chunk)
                    elif method == "$sandbox.stream.end":
                        stream_id = params.get("streamId")
                        stream = self._pending_processes.get(stream_id)
                        if stream:
                            stream.feed_eof()
                            del self._pending_processes[stream_id]

        except ConnectionClosed:
            # Cancel all pending requests when connection closes
            for future in self._pending_requests.values():
                if not future.done():
                    future.cancel()
            self._pending_requests.clear()
        except Exception as e:
            for future in self._pending_requests.values():
                if not future.done():
                    future.set_exception(e)

    async def fetch(
        self,
        url: str,
        method: Optional[str] = "GET",
        headers: Optional[dict[str, str]] = None,
        redirect: Optional[Literal["follow", "manual"]] = None,
        pid: Optional[int] = None,
    ) -> AsyncFetchResponse:
        self._signal_id += 1
        signal_id = self._signal_id

        params: FetchParams = {
            "url": url,
            "method": method or "GET",
            "headers": list(headers.items()) if headers else [],
            "redirect": redirect or "follow",
            "abortId": signal_id,
        }

        if pid is not None:
            params["pid"] = pid

        response_data = await self.call("fetch", params)
        response = cast(FetchResponseData, response_data)

        fetch_response = AsyncFetchResponse(self, response)
        return fetch_response


class FetchParams(TypedDict):
    method: str
    url: str
    headers: list[tuple[str, str]]
    redirect: str
    pid: NotRequired[int | None]
    abortId: int


class FetchResponseData(TypedDict):
    status: int
    status_text: str
    headers: list[tuple[str, str]]
    body_stream_id: int


class AsyncFetchResponse:
    def __init__(self, rpc: AsyncRpcClient, response: FetchResponseData):
        self._rpc = rpc
        self._response = response

    def raise_for_status(self) -> HTTPStatusError | None:
        if self.is_success:
            return

        message = "{self} resulted in a {error_type} (status code: {self.status_code})"
        status_class = self.status_code // 100
        error_types = {
            1: "Informational response",
            3: "Redirect response",
            4: "Client error",
            5: "Server error",
        }
        error_type = error_types.get(status_class, "Invalid status code")
        message = message.format(self, error_type=error_type)

        return HTTPStatusError(self.status_code, message)

    @property
    def headers(self) -> list[tuple[str, str]]:
        return self._response["headers"]

    @property
    def status_code(self) -> int:
        return self._response["status"]

    @property
    def is_informational(self) -> int:
        return 100 <= self.status_code <= 199

    @property
    def is_success(self) -> int:
        return 200 <= self.status_code <= 299

    @property
    def is_redirect(self) -> int:
        return 300 <= self.status_code <= 399

    @property
    def is_client_error(self) -> int:
        return 400 <= self.status_code <= 499

    @property
    def is_server_error(self) -> int:
        return 500 <= self.status_code <= 599

    @property
    def is_error(self) -> int:
        return 400 <= self.status_code <= 599

    @property
    def has_redirect_location(self) -> bool:
        return (
            self.status_code in (301, 302, 303, 307, 308)
            and "location" in self._response["headers"]
        )

    def __repr__(self) -> str:
        return f"<Response [{self.status_code}]>"


class FetchResponse(AsyncFetchResponse):
    def __init__(self, async_res: AsyncFetchResponse):
        self._async = async_res

    def raise_for_status(self) -> HTTPStatusError | None:
        return self._async.raise_for_status()

    @property
    def headers(self) -> list[tuple[str, str]]:
        return self._async.headers

    @property
    def status_code(self) -> int:
        return self._async.status_code

    @property
    def is_informational(self) -> int:
        return self._async.is_informational

    @property
    def is_success(self) -> int:
        return self._async.is_success

    @property
    def is_redirect(self) -> int:
        return self._async.is_redirect

    @property
    def is_client_error(self) -> int:
        return self._async.is_client_error

    @property
    def is_server_error(self) -> int:
        return self._async.is_server_error

    @property
    def is_error(self) -> int:
        return self._async.is_error

    @property
    def has_redirect_location(self) -> bool:
        return self._async.has_redirect_location

    def __repr__(self) -> str:
        return f"<Response [{self.status_code}]>"



from __future__ import annotations

from typing import Any, TypedDict, cast
from typing_extensions import Optional

from .bridge import AsyncBridge
from .console import (
    AsyncConsoleClient,
    AsyncPaginatedList,
    PaginatedList,
)

from .utils import convert_to_snake_case


class BaseSnapshot(TypedDict):
    id: str
    """The unique identifier for the snapshot."""

    slug: str
    """Human readable identifier for the snapshot."""


class Snapshot(TypedDict):
    id: str
    """The unique identifier for the snapshot."""

    slug: str
    """The human readable identifier for the snapshot."""

    region: str
    """The region the snapshot is located in."""

    allocated_size: int
    """
  The number of bytes currently allocated specifically for the snapshot.

  Snapshots created from other snapshots only store data that is different
  from their base snapshot, so this size may be smaller than the actual size
  of the file system on the snapshot.
  This is not the total size of the snapshot.
  """

    flattened_size: int
    """
  The total size of the file system on the snapshot, in bytes. This is the size the snapshot would take up if it were fully "flattened" (i.e., if all data from any snapshots it was created from were fully stored in this snapshot).

  Snapshots created from other snapshots only store data that is different from their base snapshot, so this size may be larger than the amount of data actually allocated for the snapshot.
  """

    is_bootable: bool
    """Whether the snapshot is bootable."""

    base_snapshot: BaseSnapshot
    """The snapshot the volume was created from."""


class AsyncSnapshots:
    def __init__(self, client: AsyncConsoleClient):
        self._client = client

    async def get(self, id_or_slug: str) -> Snapshot | None:
        """Get a snapshot by ID or slug."""
        result = await self._client.get_or_none(f"/api/v2/snapshots/{id_or_slug}")
        if result is None:
            return None
        raw_result = convert_to_snake_case(result)
        return cast(Snapshot, raw_result)

    async def list(
        self,
        *,
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
        search: Optional[str] = None,
    ) -> AsyncPaginatedList[Snapshot]:
        """List snapshots.

        Args:
            cursor: The cursor for pagination.
            limit: Limit the number of snapshots to return.
            search: The search query to filter snapshots by.
        """
        options: dict[str, Any] = {}
        if cursor is not None:
            options["cursor"] = cursor
        if limit is not None:
            options["limit"] = limit
        if search is not None:
            options["search"] = search
        return await self._client.get_paginated(
            "/api/v2/snapshots", cursor=None, params=options if options else None
        )

    async def delete(self, id_or_slug: str) -> None:
        """Delete snapshot by ID or slug."""
        await self._client.delete(f"/api/v2/snapshots/{id_or_slug}")


class Snapshots:
    def __init__(self, client: AsyncConsoleClient, bridge: AsyncBridge):
        self._client = client
        self._bridge = bridge
        self._async = AsyncSnapshots(client)

    def get(self, id_or_slug: str) -> Snapshot | None:
        """Get a snapshot by ID or slug."""
        return self._bridge.run(self._async.get(id_or_slug))

    def list(
        self,
        *,
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
        search: Optional[str] = None,
    ) -> PaginatedList[Snapshot]:
        """List snapshots.

        Args:
            cursor: The cursor for pagination.
            limit: Limit the number of snapshots to return.
            search: The search query to filter snapshots by.
        """
        paginated = self._bridge.run(
            self._async.list(cursor=cursor, limit=limit, search=search)
        )
        return PaginatedList(self._bridge, paginated)

    def delete(self, id_or_slug: str) -> None:
        """Delete snapshot by ID or slug."""
        self._bridge.run(self._async.delete(id_or_slug))



from typing import Optional

from .apps import Apps, AsyncApps
from .revisions import Revisions, AsyncRevisions
from .snapshots import Snapshots, AsyncSnapshots
from .timelines import Timelines, AsyncTimelines
from .volumes import Volumes, AsyncVolumes
from .bridge import AsyncBridge
from .sandbox import (
    AsyncSandboxApi,
    SandboxApi,
)
from .console import AsyncConsoleClient
from .options import Options, get_internal_options

__all__ = ["DenoDeploy", "AsyncDenoDeploy", "Options"]


class DenoDeploy:
    def __init__(self, options: Optional[Options] = None):
        internal_options = get_internal_options(options)
        bridge = AsyncBridge()

        client = AsyncConsoleClient(internal_options)
        self.apps = Apps(client, bridge)
        self.revisions = Revisions(client, bridge)
        self.timelines = Timelines(client, bridge)
        self.sandbox = SandboxApi(client, bridge)
        self.snapshots = Snapshots(client, bridge)
        self.volumes = Volumes(client, bridge)


class AsyncDenoDeploy:
    def __init__(self, options: Optional[Options] = None):
        internal_options = get_internal_options(options)
        client = AsyncConsoleClient(internal_options)

        self.apps = AsyncApps(client)
        self.revisions = AsyncRevisions(client)
        self.timelines = AsyncTimelines(client)
        self.sandbox = AsyncSandboxApi(client)
        self.snapshots = AsyncSnapshots(client)
        self.volumes = AsyncVolumes(client)



from __future__ import annotations

from typing_extensions import NotRequired
from typing import (
    Any,
    Generic,
    Literal,
    Optional,
    TypedDict,
    TypeVar,
    cast,
)
import httpx

from .bridge import AsyncBridge
from .options import InternalOptions
from .utils import convert_to_snake_case, parse_link_header

T = TypeVar("T")


class ExposeSSHResult(TypedDict):
    hostname: str
    username: str
    port: int


class PaginationOptions(TypedDict):
    cursor: NotRequired[str | None]
    """The cursor for pagination."""

    limit: NotRequired[int | None]
    """Limit the number of items to return."""


class AsyncPaginatedList(Generic[T]):
    def __init__(
        self,
        client: AsyncConsoleClient,
        items: list[T],
        path: str,
        next_cursor: str | None,
        options: Optional[dict[str, Any]] = None,
    ):
        self._client = client
        self._options = options
        self._path = path

        self.items = items
        """The items in the current page of the paginated list."""

        self.next_cursor = next_cursor
        """The cursor for pagination."""

    async def get_next_page(self) -> AsyncPaginatedList[T] | None:
        if not self.has_more:
            return None

        return await self._client.get_paginated(
            self._path,
            self.next_cursor,
            self._options,
        )

    @property
    def has_more(self) -> bool:
        """Whether there are more items to fetch."""
        return self.next_cursor is not None

    def __aiter__(self):
        return self

    async def __anext__(self):
        if not self.has_more:
            raise StopAsyncIteration

        next_page = await self.get_next_page()
        if next_page is None:
            raise StopAsyncIteration

        return next_page


class PaginatedList(Generic[T]):
    def __init__(
        self,
        bridge: AsyncBridge,
        async_paginated: AsyncPaginatedList[T],
    ):
        self._bridge = bridge
        self._async_paginated = async_paginated

    def get_next_page(self) -> PaginatedList[T] | None:
        next_page: AsyncPaginatedList[T] | None = self._bridge.run(
            self._async_paginated.get_next_page()
        )
        if next_page is None:
            return None

        return PaginatedList(self._bridge, next_page)

    @property
    def items(self) -> list[T]:
        """The items in the current page of the paginated list."""
        return self._async_paginated.items

    @property
    def next_cursor(self) -> str | None:
        """The cursor for pagination."""
        return self._async_paginated.next_cursor

    @property
    def has_more(self) -> bool:
        """Whether there are more items to fetch."""
        return self._async_paginated.has_more


class AsyncConsoleClient:
    def __init__(self, options: InternalOptions):
        self._options = options
        self._client: httpx.AsyncClient | None = None

    @property
    def client(self) -> httpx.AsyncClient:
        if self._client is None:
            self._client = httpx.AsyncClient(
                headers={
                    "Content-Type": "application/json",
                    "Authorization": f"Bearer {self._options['token']}",
                }
            )
        return self._client

    async def _request(
        self,
        method: Literal["POST", "GET", "PATCH", "PUT", "DELETE"],
        url: httpx.URL,
        data: Optional[Any] = None,
    ) -> httpx.Response:
        response = await self.client.request(
            method=method, url=url, json=data, timeout=10.0
        )

        response.raise_for_status()
        return response

    async def post(self, path: str, data: Any) -> dict:
        req_url = self._options["console_url"].join(path)
        response = await self._request("POST", req_url, data)
        return response.json()

    async def patch(self, path: str, data: Any) -> dict:
        req_url = self._options["console_url"].join(path)
        response = await self._request("PATCH", req_url, data)
        return response.json()

    async def get(
        self, path: str, params: Optional[dict[str, str | int]] = None
    ) -> dict:
        req_url = self._options["console_url"].join(path)

        if params is not None:
            req_url = req_url.copy_merge_params(params)

        response = await self._request("GET", req_url)
        return response.json()

    async def get_or_none(
        self, path: str, params: Optional[dict[str, str | int]] = None
    ) -> dict | None:
        try:
            return await self.get(path, params)
        except httpx.HTTPStatusError as e:
            if e.response.status_code == 404:
                return None
            raise

    async def delete(self, path: str) -> httpx.Response:
        req_url = self._options["console_url"].join(path)
        response = await self._request("DELETE", req_url)
        return response

    async def get_paginated(
        self,
        path: str,
        cursor: Optional[str],
        params: Optional[dict[str, Any]] = None,
    ) -> AsyncPaginatedList[T]:
        req_url = self._options["console_url"].join(path)

        if params is not None:
            url_params = cast(dict, params)
            req_url = req_url.copy_merge_params(url_params)

        if cursor is not None:
            req_url = req_url.copy_add_param("cursor", cursor)

        response = await self._request("GET", req_url)
        response.raise_for_status()
        data = response.json()

        next_cursor: str | None = None
        link_header = response.headers.get("link")
        if link_header is not None:
            parsed = parse_link_header(link_header)
            next_cursor = parsed.get("next", None)

        items = [cast(T, convert_to_snake_case(item)) for item in data]

        return AsyncPaginatedList(self, items, path, next_cursor, params)

    async def close(self) -> None:
        await self.client.aclose()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()


__all__ = ["AsyncConsoleClient"]



from __future__ import annotations

from typing import Any, TypedDict, Union, cast
from typing_extensions import Optional, NotRequired

from .snapshots import BaseSnapshot, Snapshot

from .utils import convert_to_snake_case

from .bridge import AsyncBridge
from .console import (
    AsyncConsoleClient,
    AsyncPaginatedList,
    PaginatedList,
)


class Volume(TypedDict):
    id: str
    """The unique identifier for the volume."""

    slug: str
    """Human readable identifier for the volume."""

    region: str
    """The region the volume is located in."""

    capacity: int
    """The capacity of the volume in bytes."""

    estimated_allocated_size: int
    """
  The number of bytes currently allocated specifically for the volume.

  Volumes created from snapshots are "copy-on-write", so this size may be
  smaller than the actual size of the file system on the volume, as it does
  not include data shared with any snapshots the volume was created from.

  This is an estimate, and may lag real-time usage by multiple minutes.
  """

    estimated_flattened_size: int
    """
  The total size of the file system on the volume, in bytes. This is the size
  the volume would take up if it were fully "flattened" (i.e., if all data
  from any snapshots it was created from were fully copied to the volume).

  Volumes created from snapshots are "copy-on-write", so this size may be
  larger than the amount of data actually allocated for the volume.

  This is an estimate, and may lag real-time usage by multiple minutes.
  """

    is_bootable: bool
    """Whether the volume is bootable."""

    base_snapshot: NotRequired[BaseSnapshot | None]
    """The snapshot the volume was created from."""


class AsyncVolumes:
    def __init__(self, client: AsyncConsoleClient):
        self._client = client

    async def create(
        self,
        slug: str,
        region: str,
        capacity: Union[int, str],
        *,
        from_snapshot: Optional[str] = None,
    ) -> Volume:
        """Create a new volume.

        Args:
            slug: Human readable identifier for the volume.
            region: The region to create the volume in.
            capacity: The capacity of the volume in bytes. When passing a string you can use these units too: GB, MB, kB, GiB, MiB, KiB.
            from_snapshot: The ID or slug of the snapshot to create the volume from.
        """
        params: dict[str, Any] = {
            "slug": slug,
            "capacity": capacity,
            "region": region,
        }
        if from_snapshot is not None:
            params["from"] = from_snapshot

        result = await self._client.post("/api/v2/volumes", params)
        raw_result = convert_to_snake_case(result)
        return cast(Volume, raw_result)

    async def get(self, id_or_slug: str) -> Volume | None:
        """Get volume info by ID or slug."""
        result = await self._client.get_or_none(f"/api/v2/volumes/{id_or_slug}")
        if result is None:
            return None
        raw_result = convert_to_snake_case(result)
        return cast(Volume, raw_result)

    async def list(
        self,
        *,
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
        search: Optional[str] = None,
    ) -> AsyncPaginatedList[Volume]:
        """List volumes.

        Args:
            cursor: The cursor for pagination.
            limit: Limit the number of volumes to return.
            search: The search query to filter volumes by.
        """
        options: dict[str, Any] = {}
        if cursor is not None:
            options["cursor"] = cursor
        if limit is not None:
            options["limit"] = limit
        if search is not None:
            options["search"] = search

        result: AsyncPaginatedList[Volume] = await self._client.get_paginated(
            "/api/v2/volumes", cursor=None, params=options if options else None
        )

        return result

    async def delete(self, id_or_slug: str) -> None:
        """Delete volume by ID or slug."""
        await self._client.delete(f"/api/v2/volumes/{id_or_slug}")

    async def snapshot(
        self,
        id_or_slug: str,
        *,
        slug: str,
    ) -> Snapshot:
        """Create a snapshot of a volume by ID or slug.

        Args:
            id_or_slug: The volume ID or slug to snapshot.
            slug: Human readable identifier for the snapshot.
        """
        params: dict[str, Any] = {"slug": slug}
        result = await self._client.post(
            f"/api/v2/volumes/{id_or_slug}/snapshot", params
        )
        raw_result = convert_to_snake_case(result)
        return cast(Snapshot, raw_result)


class Volumes:
    def __init__(self, client: AsyncConsoleClient, bridge: AsyncBridge):
        self._client = client
        self._bridge = bridge
        self._async = AsyncVolumes(client)

    def create(
        self,
        slug: str,
        region: str,
        capacity: Union[int, str],
        *,
        from_snapshot: Optional[str] = None,
    ) -> Volume:
        """Create a new volume.

        Args:
            slug: Human readable identifier for the volume.
            region: The region to create the volume in.
            capacity: The capacity of the volume in bytes. When passing a string you can use these units too: GB, MB, kB, GiB, MiB, KiB.
            from_snapshot: The ID or slug of the snapshot to create the volume from.
        """
        return self._bridge.run(
            self._async.create(slug, region, capacity, from_snapshot=from_snapshot)
        )

    def get(self, id_or_slug: str) -> Volume | None:
        """Get volume info by ID or slug."""
        return self._bridge.run(self._async.get(id_or_slug))

    def list(
        self,
        *,
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
        search: Optional[str] = None,
    ) -> PaginatedList[Volume]:
        """List volumes.

        Args:
            cursor: The cursor for pagination.
            limit: Limit the number of volumes to return.
            search: The search query to filter volumes by.
        """
        paginated = self._bridge.run(
            self._async.list(cursor=cursor, limit=limit, search=search)
        )
        return PaginatedList(self._bridge, paginated)

    def delete(self, id_or_slug: str) -> None:
        """Delete volume by ID or slug."""
        self._bridge.run(self._async.delete(id_or_slug))

    def snapshot(
        self,
        id_or_slug: str,
        *,
        slug: str,
    ) -> Snapshot:
        """Create a snapshot of a volume by ID or slug.

        Args:
            id_or_slug: The volume ID or slug to snapshot.
            slug: Human readable identifier for the snapshot.
        """
        return self._bridge.run(self._async.snapshot(id_or_slug, slug=slug))



from __future__ import annotations

import base64
import os
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterable,
    BinaryIO,
    Iterable,
    Literal,
    Optional,
    TypedDict,
    Union,
    cast,
)

from re import Pattern

from .process import AbortSignal
from .stream import stream_data
from .utils import convert_to_camel_case, convert_to_snake_case

if TYPE_CHECKING:
    from .rpc import AsyncRpcClient
    from .bridge import AsyncBridge


class DirEntry(TypedDict):
    name: str
    """The file or directory name."""

    is_file: bool
    """Whether the entry is a file."""

    is_directory: bool
    """Whether the entry is a directory."""

    is_symlink: bool
    """Whether the entry is a symbolic link."""


class FileInfo(TypedDict):
    is_file: bool
    """Whether the path is a file."""

    is_directory: bool
    """Whether the path is a directory."""

    is_symlink: bool
    """Whether the path is a symbolic link."""

    size: int
    """The size of the file in bytes."""

    mtime: str
    """The last modification time of the file."""

    atime: str
    """The last access time of the file."""

    birthtime: str
    """The creation time of the file."""

    ctime: str
    """The last status change time of the file."""

    dev: int
    """The device ID."""

    ino: int
    """The inode number."""

    mode: int
    """The file mode."""

    nlink: int
    """The number of hard links pointing to this file."""

    uid: int
    """The user ID of the owner."""

    gid: int
    """The group ID of the owner."""

    rdev: int
    """The device ID of this file."""

    blksize: int
    """The block size for filesystem I/O."""

    blocks: int
    """The number of blocks allocated in 512-byte units"""

    is_block_device: bool
    """Whether the path is a block device."""

    is_char_device: bool
    """Whether the path is a character device."""

    is_fifo: bool
    """Whether the path is a FIFO."""

    is_socket: bool
    """Whether the path is a socket."""


class WalkEntry(TypedDict):
    path: str
    """The full path of the entry."""

    name: str
    """The file or directory name."""

    is_file: bool
    """Whether the entry is a file."""

    is_directory: bool
    """Whether the entry is a directory."""

    is_symlink: bool
    """Whether the entry is a symbolic link."""


class FsFileHandle(TypedDict):
    file_handle_id: int


class AsyncFsFile:
    def __init__(self, rpc: AsyncRpcClient, fd: int):
        self._rpc = rpc
        self._fd = fd

    async def write(self, data: bytes) -> int:
        """Write data to the file. Returns number of bytes written."""

        result = await self._rpc.call(
            "fileWrite",
            {"data": base64.b64encode(data).decode("ascii"), "fileHandleId": self._fd},
        )
        return result["bytes_written"]

    async def truncate(self, size: Optional[int]) -> None:
        """Truncate the file to the given size. If size is None, truncate to 0."""

        await self._rpc.call("fileTruncate", {"size": size, "fileHandleId": self._fd})

    async def read(self, size: int) -> bytes:
        """Read up to size bytes from the file. Returns the data read."""

        result = await self._rpc.call(
            "fileRead", {"length": size, "fileHandleId": self._fd}
        )
        return base64.b64decode(result["data"])

    async def seek(self, offset: int, whence: int) -> int:
        """Seek to a position in the file. Returns the new position."""

        result = await self._rpc.call(
            "fileSeek", {"offset": offset, "whence": whence, "fileHandleId": self._fd}
        )
        return result["position"]

    async def stat(self) -> FileInfo:
        """Get file information."""

        result = await self._rpc.call("fileStat", {"fileHandleId": self._fd})
        return cast(FileInfo, result)

    async def sync(self) -> None:
        """Flushes any pending data and metadata operations of the given file stream to disk."""

        await self._rpc.call("fileSync", {"fileHandleId": self._fd})

    async def sync_data(self) -> None:
        """Sync the file's data to disk."""

        await self._rpc.call("fileSyncData", {"fileHandleId": self._fd})

    async def utime(self, atime: float, mtime: float) -> None:
        """Update the file's access and modification times."""

        await self._rpc.call(
            "fileUtime",
            {"atime": atime, "mtime": mtime, "fileHandleId": self._fd},
        )

    async def lock(self, exclusive: Optional[bool]) -> None:
        """Lock the file. If exclusive is True, acquire an exclusive lock."""

        await self._rpc.call(
            "fileLock", {"exclusive": exclusive, "fileHandleId": self._fd}
        )

    async def unlock(self) -> None:
        """Unlock the file."""

        await self._rpc.call("fileUnlock", {"fileHandleId": self._fd})

    async def close(self) -> None:
        """Close the file."""

        await self._rpc.call("fileClose", {"fileHandleId": self._fd})

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()


class FsFile:
    def __init__(self, rpc: AsyncRpcClient, bridge: AsyncBridge, fd: int):
        self._rpc = rpc
        self._bridge = bridge
        self._async = AsyncFsFile(rpc, fd)

    def write(self, data: bytes) -> int:
        """Write data to the file. Returns number of bytes written."""
        return self._bridge.run(self._async.write(data))

    def truncate(self, size: Optional[int]) -> None:
        """Truncate the file to the given size. If size is None, truncate to 0."""
        return self._bridge.run(self._async.truncate(size))

    def read(self, size: int) -> bytes:
        """Read up to size bytes from the file. Returns the data read."""
        return self._bridge.run(self._async.read(size))

    def seek(self, offset: int, whence: int) -> int:
        """Seek to a position in the file. Returns the new position."""
        return self._bridge.run(self._async.seek(offset, whence))

    def stat(self) -> FileInfo:
        """Get file information."""
        return self._bridge.run(self._async.stat())

    def sync(self) -> None:
        """Flushes any pending data and metadata operations of the given file stream to disk."""
        return self._bridge.run(self._async.sync())

    def sync_data(self) -> None:
        """Sync the file's data to disk."""
        return self._bridge.run(self._async.sync_data())

    def utime(self, atime: int, mtime: int) -> None:
        """Update the file's access and modification times."""
        return self._bridge.run(self._async.utime(atime, mtime))

    def lock(self, exclusive: Optional[bool]) -> None:
        """Lock the file. If exclusive is True, acquire an exclusive lock."""
        return self._bridge.run(self._async.lock(exclusive))

    def unlock(self) -> None:
        """Unlock the file."""
        return self._bridge.run(self._async.unlock())

    def close(self) -> None:
        """Close the file."""
        return self._bridge.run(self._async.close())

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class AsyncSandboxFs:
    """Filesystem operations inside the sandbox."""

    def __init__(self, rpc: AsyncRpcClient):
        self._rpc = rpc

    async def read_file(
        self,
        path: str,
        *,
        signal: Optional[AbortSignal] = None,
    ) -> bytes:
        """Reads the entire contents of a file as bytes.

        Args:
            path: The path to the file to read.
            signal: An optional abort signal to cancel the operation.
        """
        params: dict[str, Any] = {"path": path}
        options: dict[str, Any] = {}
        if signal is not None:
            options["signal"] = signal
        if options:
            params["options"] = convert_to_camel_case(options)

        result = await self._rpc.call("readFile", params)

        # Server returns base64-encoded data
        return base64.b64decode(result)

    async def write_file(
        self,
        path: str,
        data: Union[bytes, AsyncIterable[bytes], Iterable[bytes], BinaryIO],
        *,
        create: Optional[bool] = None,
        append: Optional[bool] = None,
        create_new: Optional[bool] = None,
        mode: Optional[int] = None,
    ) -> None:
        """Write bytes to file. Accepts bytes, async/sync iterables, or file objects.

        Args:
            path: The path to the file to write.
            data: The data to write to the file.
            create: Create file if it doesn't already exist.
            append: Append content instead of overwriting existing contents.
            create_new: Fail if the file already exists.
            mode: Set the file permission mode.
        """
        if isinstance(data, bytes):
            # Stream bytes as a single chunk
            content_stream_id = await stream_data(self._rpc, iter([data]))
        else:
            # Stream data from iterable/file object
            content_stream_id = await stream_data(self._rpc, data)

        params: dict[str, Any] = {"path": path, "contentStreamId": content_stream_id}
        options: dict[str, Any] = {}
        if create is not None:
            options["create"] = create
        if append is not None:
            options["append"] = append
        if create_new is not None:
            options["create_new"] = create_new
        if mode is not None:
            options["mode"] = mode
        if options:
            params["options"] = convert_to_camel_case(options)
        await self._rpc.call("writeFile", params)

    async def read_text_file(
        self,
        path: str,
        *,
        signal: Optional[AbortSignal] = None,
    ) -> str:
        """Reads the entire contents of a file as an UTF-8 decoded string.

        Args:
            path: The path to the file to read.
            signal: An optional abort signal to cancel the operation.
        """
        params: dict[str, Any] = {"path": path}
        options: dict[str, Any] = {}
        if signal is not None:
            options["signal"] = signal
        if options:
            params["options"] = convert_to_camel_case(options)

        result = await self._rpc.call("readTextFile", params)

        return result

    async def write_text_file(
        self,
        path: str,
        content: str,
        *,
        create: Optional[bool] = None,
        append: Optional[bool] = None,
        create_new: Optional[bool] = None,
        mode: Optional[int] = None,
    ) -> None:
        """Write text to file. Creates a new file if needed. Existing files will be overwritten.

        Args:
            path: The path to the file to write.
            content: The text content to write.
            create: Create file if it doesn't already exist.
            append: Append content instead of overwriting existing contents.
            create_new: Fail if the file already exists.
            mode: Set the file permission mode.
        """
        params: dict[str, Any] = {"path": path, "content": content}
        options: dict[str, Any] = {}
        if create is not None:
            options["create"] = create
        if append is not None:
            options["append"] = append
        if create_new is not None:
            options["create_new"] = create_new
        if mode is not None:
            options["mode"] = mode
        if options:
            params["options"] = convert_to_camel_case(options)

        await self._rpc.call("writeTextFile", params)

    async def read_dir(self, path: str) -> list[DirEntry]:
        """Read the directory entries at the given path."""

        params = {"path": path}
        result = await self._rpc.call("readDir", params)

        return result

    async def remove(
        self,
        path: str,
        *,
        recursive: Optional[bool] = None,
    ) -> None:
        """Remove the file or directory at the given path.

        Args:
            path: The path to the file or directory to remove.
            recursive: Whether to remove directories recursively. Default: false.
        """
        params: dict[str, Any] = {"path": path}
        options: dict[str, Any] = {}
        if recursive is not None:
            options["recursive"] = recursive
        if options:
            params["options"] = convert_to_camel_case(options)

        await self._rpc.call("remove", params)

    async def mkdir(
        self,
        path: str,
        *,
        recursive: Optional[bool] = None,
        mode: Optional[int] = None,
    ) -> None:
        """Create a new directory at the specified path.

        Args:
            path: The path to the directory to create.
            recursive: If true, parent directories will be created if they do not already exist.
            mode: Permissions to use for the new directory.
        """
        params: dict[str, Any] = {"path": path}
        options: dict[str, Any] = {}
        if recursive is not None:
            options["recursive"] = recursive
        if mode is not None:
            options["mode"] = mode
        if options:
            params["options"] = convert_to_camel_case(options)

        await self._rpc.call("mkdir", params)

    async def rename(self, old_path: str, new_path: str) -> None:
        """Rename (move) a file or directory."""

        params = {"oldPath": old_path, "newPath": new_path}
        await self._rpc.call("rename", params)

    async def stat(self, path: str) -> FileInfo:
        """Return file information about a file or directory."""

        params = {"path": path}
        result = await self._rpc.call("stat", params)

        raw_result = convert_to_snake_case(result)
        return cast(FileInfo, raw_result)

    async def chmod(self, path: str, mode: int) -> None:
        """Change the permission mode of a file or directory."""

        params = {"path": path, "mode": mode}
        await self._rpc.call("chmod", params)

    async def chown(
        self, path: str, uid: Optional[int] = None, gid: Optional[int] = None
    ) -> None:
        """Change the owner user ID and group ID of a file or directory."""

        params = {"path": path, "uid": uid, "gid": gid}
        await self._rpc.call("chown", params)

    async def copy_file(self, from_path: str, to_path: str) -> None:
        """Copy a file from a source path to a destination path."""

        params = {"fromPath": from_path, "toPath": to_path}
        await self._rpc.call("copyFile", params)

    async def walk(
        self,
        path: str,
        *,
        max_depth: Optional[int] = None,
        include_files: Optional[bool] = None,
        include_dirs: Optional[bool] = None,
        include_symlinks: Optional[bool] = None,
        follow_symlinks: Optional[bool] = None,
        canonicalize: Optional[bool] = None,
        exts: Optional[list[str]] = None,
        match: Optional[list[Pattern]] = None,
        skip: Optional[list[Pattern]] = None,
    ) -> list[WalkEntry]:
        """Recursively walk a directory tree.

        Args:
            path: The path to the directory to walk.
            max_depth: The maximum depth to traverse. Default: Infinity.
            include_files: Whether to include files in the results. Default: true.
            include_dirs: Whether to include directories in the results. Default: true.
            include_symlinks: Whether to include symbolic links in the results. Default: true.
            follow_symlinks: Whether to follow symbolic links. Default: false.
            canonicalize: Whether to return canonicalized paths. This option works only if `follow_symlinks` is not `false`. Default: false.
            exts: If provided, only files with the specified extensions will be included. Example: ['.ts', '.js']
            match: List of regular expression patterns used to filter entries. If specified, entries that do not match the patterns specified by this option are excluded.
            skip: List of regular expression patterns used to filter entries. If specified, entries that match the patterns specified by this option are excluded.
        """
        params: dict[str, Any] = {"path": path}
        options: dict[str, Any] = {}
        if max_depth is not None:
            options["max_depth"] = max_depth
        if include_files is not None:
            options["include_files"] = include_files
        if include_dirs is not None:
            options["include_dirs"] = include_dirs
        if include_symlinks is not None:
            options["include_symlinks"] = include_symlinks
        if follow_symlinks is not None:
            options["follow_symlinks"] = follow_symlinks
        if canonicalize is not None:
            options["canonicalize"] = canonicalize
        if exts is not None:
            options["exts"] = exts
        if match is not None:
            options["match"] = match
        if skip is not None:
            options["skip"] = skip
        if options:
            params["options"] = convert_to_camel_case(options)

        result = await self._rpc.call("walk", params)

        return result

    async def expand_glob(
        self,
        glob: str,
        *,
        root: Optional[str] = None,
        exclude: Optional[list[str]] = None,
        include_dirs: Optional[bool] = None,
        follow_symlinks: Optional[bool] = None,
        canonicalize: Optional[bool] = None,
        extended: Optional[bool] = None,
        globstar: Optional[bool] = None,
        case_insensitive: Optional[bool] = None,
    ) -> list[str]:
        """Expand a glob pattern to a list of paths.

        Args:
            glob: The glob pattern to expand.
            root: The root directory from which to expand the glob pattern. Default is the current working directory.
            exclude: An array of glob patterns to exclude from the results.
            include_dirs: Whether to include directories in the results. Default: true.
            follow_symlinks: Whether to follow symbolic links. Default: false.
            canonicalize: Whether to return canonicalized paths. This option works only if `follow_symlinks` is not `false`. Default: true.
            extended: Whether to enable extended glob syntax, see https://www.linuxjournal.com/content/bash-extended-globbing. Default: true.
            globstar: Globstar syntax. See https://www.linuxjournal.com/content/globstar-new-bash-globbing-option. If false, `**` is treated like `*`. Default: true.
            case_insensitive: Whether the glob matching should be case insensitive. Default: false.
        """
        params: dict[str, Any] = {"glob": glob}
        options: dict[str, Any] = {}
        if root is not None:
            options["root"] = root
        if exclude is not None:
            options["exclude"] = exclude
        if include_dirs is not None:
            options["include_dirs"] = include_dirs
        if follow_symlinks is not None:
            options["follow_symlinks"] = follow_symlinks
        if canonicalize is not None:
            options["canonicalize"] = canonicalize
        if extended is not None:
            options["extended"] = extended
        if globstar is not None:
            options["globstar"] = globstar
        if case_insensitive is not None:
            options["case_insensitive"] = case_insensitive
        if options:
            params["options"] = convert_to_camel_case(options)

        result = await self._rpc.call("expandGlob", params)

        return result

    async def link(self, target: str, path: str) -> None:
        """Create a hard link pointing to an existing file."""

        params = {"target": target, "path": path}
        await self._rpc.call("link", params)

    async def lstat(self, path: str) -> FileInfo:
        """Return file information about a file or directory symlink."""

        params = {"path": path}
        result = await self._rpc.call("lstat", params)

        raw_result = convert_to_snake_case(result)
        return cast(FileInfo, raw_result)

    async def make_temp_dir(
        self,
        *,
        dir: Optional[str] = None,
        prefix: Optional[str] = None,
        suffix: Optional[str] = None,
    ) -> str:
        """Create a new temporary directory.

        Args:
            dir: The directory where the temporary directory should be created.
            prefix: The prefix for the temporary directory name.
            suffix: The suffix for the temporary directory name.
        """
        params: dict[str, Any] = {}
        options: dict[str, Any] = {}
        if dir is not None:
            options["dir"] = dir
        if prefix is not None:
            options["prefix"] = prefix
        if suffix is not None:
            options["suffix"] = suffix
        if options:
            params["options"] = convert_to_camel_case(options)

        result = await self._rpc.call("makeTempDir", params)

        return result

    async def make_temp_file(
        self,
        *,
        dir: Optional[str] = None,
        prefix: Optional[str] = None,
        suffix: Optional[str] = None,
    ) -> str:
        """Create a new temporary file.

        Args:
            dir: The directory where the temporary file should be created.
            prefix: The prefix for the temporary file name.
            suffix: The suffix for the temporary file name.
        """
        params: dict[str, Any] = {}
        options: dict[str, Any] = {}
        if dir is not None:
            options["dir"] = dir
        if prefix is not None:
            options["prefix"] = prefix
        if suffix is not None:
            options["suffix"] = suffix
        if options:
            params["options"] = convert_to_camel_case(options)

        result = await self._rpc.call("makeTempFile", params)

        return result

    async def read_link(self, path: str) -> str:
        """Read the target of a symbolic link."""

        params = {"path": path}
        result = await self._rpc.call("readLink", params)

        return result

    async def real_path(self, path: str) -> str:
        """Return the canonicalized absolute pathname."""

        params = {"path": path}
        result = await self._rpc.call("realPath", params)

        return result

    async def symlink(
        self,
        target: str,
        path: str,
        *,
        type: Optional[Literal["file", "dir", "junction"]] = None,
    ) -> None:
        """Create a symbolic link.

        Args:
            target: The target path of the symbolic link.
            path: The path where the symbolic link should be created.
            type: The type of the symbolic link (file, dir, or junction).
        """
        params: dict[str, Any] = {"target": target, "path": path}
        options: dict[str, Any] = {}
        if type is not None:
            options["type"] = type
        if options:
            params["options"] = convert_to_camel_case(options)

        await self._rpc.call("symlink", params)

    async def truncate(self, name: str, length: Optional[int] = None) -> None:
        """Truncate or extend the specified file to reach a given size."""

        params = {"name": name, "length": length}
        await self._rpc.call("truncate", params)

    async def umask(self, mask: Optional[int] = None) -> int:
        """Sets the process's file mode creation mask."""

        params = {"mask": mask}
        result = await self._rpc.call("umask", params)

        return result

    async def utime(self, path: str, atime: str, mtime: str) -> None:
        """Change the access and modification times of a file."""

        params = {"path": path, "atime": atime, "mtime": mtime}
        await self._rpc.call("utime", params)

    async def upload(self, local_path: str, sandbox_path: str) -> None:
        """Upload a file, directory, or symlink from local filesystem to the sandbox.

        Recursively uploads directories and their contents.
        Preserves symlinks by creating corresponding symlinks in the sandbox.
        """
        await self._upload_item(local_path, sandbox_path)

    async def _upload_item(self, local_path: str, sandbox_path: str) -> None:
        """Internal method to upload a single item (file, directory, or symlink)."""
        if os.path.islink(local_path):
            # It's a symlink - read the target and create a symlink in sandbox
            target = os.readlink(local_path)
            await self.symlink(target, sandbox_path)
        elif os.path.isdir(local_path):
            # It's a directory - create it and recursively upload contents
            await self.mkdir(sandbox_path, recursive=True)
            for entry in os.listdir(local_path):
                entry_local_path = os.path.join(local_path, entry)
                entry_sandbox_path = f"{sandbox_path}/{entry}"
                await self._upload_item(entry_local_path, entry_sandbox_path)
        elif os.path.isfile(local_path):
            # It's a file - stream it to write_file
            with open(local_path, "rb") as f:
                await self.write_file(sandbox_path, f)
        else:
            raise FileNotFoundError(f"Local path does not exist: {local_path}")

    async def download(self, local_path: str, sandbox_path: str) -> None:
        """Download a file or directory from the sandbox."""

        params = {"localPath": local_path, "sandboxPath": sandbox_path}
        await self._rpc.call("download", params)

    async def create(self, path: str) -> AsyncFsFile:
        """Create a new, empty file at the specified path."""

        params = {"path": path}
        result = await self._rpc.call("create", params)

        raw_result = convert_to_snake_case(result)
        handle = cast(FsFileHandle, raw_result)

        return AsyncFsFile(self._rpc, handle["file_handle_id"])

    async def open(
        self,
        path: str,
        *,
        read: Optional[bool] = None,
        write: Optional[bool] = None,
        append: Optional[bool] = None,
        truncate: Optional[bool] = None,
        create: Optional[bool] = None,
        create_new: Optional[bool] = None,
        mode: Optional[int] = None,
    ) -> AsyncFsFile:
        """Open a file and return a file descriptor.

        Args:
            path: The path to the file to open.
            read: Sets the option for read access. This option, when `true`, means that the file should be read-able if opened. Default: true.
            write: Sets the option for write access. This option, when `true`, means that the file should be write-able if opened. Default: false.
            append: Sets the option for append mode. This option, when `true`, means that writes to the file will always append to the end. Default: false.
            truncate: If `true`, and the file already exists and is a regular file, it will be truncated to length 0 when opened. Default: false.
            create: If `true`, the file will be created if it does not already exist. Default: false.
            create_new: If `true`, the file will be created if it does not already exist. Default: false.
            mode: The permission mode to use when creating the file.
        """
        params: dict[str, Any] = {"path": path}
        options: dict[str, Any] = {}
        if read is not None:
            options["read"] = read
        if write is not None:
            options["write"] = write
        if append is not None:
            options["append"] = append
        if truncate is not None:
            options["truncate"] = truncate
        if create is not None:
            options["create"] = create
        if create_new is not None:
            options["create_new"] = create_new
        if mode is not None:
            options["mode"] = mode
        if options:
            params["options"] = convert_to_camel_case(options)

        result = await self._rpc.call("open", params)

        raw_result = convert_to_snake_case(result)
        handle = cast(FsFileHandle, raw_result)

        return AsyncFsFile(self._rpc, handle["file_handle_id"])


class SandboxFs:
    """Filesystem operations inside the sandbox."""

    def __init__(self, rpc: AsyncRpcClient, bridge: AsyncBridge):
        self._rpc = rpc
        self._bridge = bridge
        self._async = AsyncSandboxFs(rpc)

    def read_file(
        self,
        path: str,
        *,
        signal: Optional[AbortSignal] = None,
    ) -> bytes:
        """Reads the entire contents of a file as bytes.

        Args:
            path: The path to the file to read.
            signal: An optional abort signal to cancel the operation.
        """
        return self._bridge.run(self._async.read_file(path, signal=signal))

    def write_file(
        self,
        path: str,
        data: Union[bytes, Iterable[bytes], BinaryIO],
        *,
        create: Optional[bool] = None,
        append: Optional[bool] = None,
        create_new: Optional[bool] = None,
        mode: Optional[int] = None,
    ) -> None:
        """Write bytes to file. Accepts bytes, sync iterables, or file objects.

        Args:
            path: The path to the file to write.
            data: The data to write to the file.
            create: Create file if it doesn't already exist.
            append: Append content instead of overwriting existing contents.
            create_new: Fail if the file already exists.
            mode: Set the file permission mode.
        """
        self._bridge.run(
            self._async.write_file(
                path,
                data,
                create=create,
                append=append,
                create_new=create_new,
                mode=mode,
            )
        )

    def read_text_file(
        self,
        path: str,
        *,
        signal: Optional[AbortSignal] = None,
    ) -> str:
        """Reads the entire contents of a file as an UTF-8 decoded string.

        Args:
            path: The path to the file to read.
            signal: An optional abort signal to cancel the operation.
        """
        return self._bridge.run(self._async.read_text_file(path, signal=signal))

    def write_text_file(
        self,
        path: str,
        content: str,
        *,
        create: Optional[bool] = None,
        append: Optional[bool] = None,
        create_new: Optional[bool] = None,
        mode: Optional[int] = None,
    ) -> None:
        """Write text to file. Creates a new file if needed. Existing files will be overwritten.

        Args:
            path: The path to the file to write.
            content: The text content to write.
            create: Create file if it doesn't already exist.
            append: Append content instead of overwriting existing contents.
            create_new: Fail if the file already exists.
            mode: Set the file permission mode.
        """
        self._bridge.run(
            self._async.write_text_file(
                path,
                content,
                create=create,
                append=append,
                create_new=create_new,
                mode=mode,
            )
        )

    def read_dir(self, path: str) -> list[DirEntry]:
        """Read the directory entries at the given path."""
        return self._bridge.run(self._async.read_dir(path))

    def remove(
        self,
        path: str,
        *,
        recursive: Optional[bool] = None,
    ) -> None:
        """Remove the file or directory at the given path.

        Args:
            path: The path to the file or directory to remove.
            recursive: Whether to remove directories recursively. Default: false.
        """
        self._bridge.run(self._async.remove(path, recursive=recursive))

    def mkdir(
        self,
        path: str,
        *,
        recursive: Optional[bool] = None,
        mode: Optional[int] = None,
    ) -> None:
        """Create a new directory at the specified path.

        Args:
            path: The path to the directory to create.
            recursive: If true, parent directories will be created if they do not already exist.
            mode: Permissions to use for the new directory.
        """
        self._bridge.run(self._async.mkdir(path, recursive=recursive, mode=mode))

    def rename(self, old_path: str, new_path: str) -> None:
        """Rename (move) a file or directory."""
        self._bridge.run(self._async.rename(old_path, new_path))

    def stat(self, path: str) -> FileInfo:
        """Return file information about a file or directory."""
        return self._bridge.run(self._async.stat(path))

    def chmod(self, path: str, mode: int) -> None:
        """Change the permission mode of a file or directory."""
        self._bridge.run(self._async.chmod(path, mode))

    def chown(
        self, path: str, uid: Optional[int] = None, gid: Optional[int] = None
    ) -> None:
        """Change the owner user ID and group ID of a file or directory."""
        self._bridge.run(self._async.chown(path, uid, gid))

    def copy_file(self, from_path: str, to_path: str) -> None:
        """Copy a file from a source path to a destination path."""
        self._bridge.run(self._async.copy_file(from_path, to_path))

    def walk(
        self,
        path: str,
        *,
        max_depth: Optional[int] = None,
        include_files: Optional[bool] = None,
        include_dirs: Optional[bool] = None,
        include_symlinks: Optional[bool] = None,
        follow_symlinks: Optional[bool] = None,
        canonicalize: Optional[bool] = None,
        exts: Optional[list[str]] = None,
        match: Optional[list[Pattern]] = None,
        skip: Optional[list[Pattern]] = None,
    ) -> list[WalkEntry]:
        """Recursively walk a directory tree.

        Args:
            path: The path to the directory to walk.
            max_depth: The maximum depth to traverse. Default: Infinity.
            include_files: Whether to include files in the results. Default: true.
            include_dirs: Whether to include directories in the results. Default: true.
            include_symlinks: Whether to include symbolic links in the results. Default: true.
            follow_symlinks: Whether to follow symbolic links. Default: false.
            canonicalize: Whether to return canonicalized paths. This option works only if `follow_symlinks` is not `false`. Default: false.
            exts: If provided, only files with the specified extensions will be included. Example: ['.ts', '.js']
            match: List of regular expression patterns used to filter entries. If specified, entries that do not match the patterns specified by this option are excluded.
            skip: List of regular expression patterns used to filter entries. If specified, entries that match the patterns specified by this option are excluded.
        """
        return self._bridge.run(
            self._async.walk(
                path,
                max_depth=max_depth,
                include_files=include_files,
                include_dirs=include_dirs,
                include_symlinks=include_symlinks,
                follow_symlinks=follow_symlinks,
                canonicalize=canonicalize,
                exts=exts,
                match=match,
                skip=skip,
            )
        )

    def expand_glob(
        self,
        glob: str,
        *,
        root: Optional[str] = None,
        exclude: Optional[list[str]] = None,
        include_dirs: Optional[bool] = None,
        follow_symlinks: Optional[bool] = None,
        canonicalize: Optional[bool] = None,
        extended: Optional[bool] = None,
        globstar: Optional[bool] = None,
        case_insensitive: Optional[bool] = None,
    ) -> list[str]:
        """Expand a glob pattern to a list of paths.

        Args:
            glob: The glob pattern to expand.
            root: The root directory from which to expand the glob pattern. Default is the current working directory.
            exclude: An array of glob patterns to exclude from the results.
            include_dirs: Whether to include directories in the results. Default: true.
            follow_symlinks: Whether to follow symbolic links. Default: false.
            canonicalize: Whether to return canonicalized paths. This option works only if `follow_symlinks` is not `false`. Default: true.
            extended: Whether to enable extended glob syntax, see https://www.linuxjournal.com/content/bash-extended-globbing. Default: true.
            globstar: Globstar syntax. See https://www.linuxjournal.com/content/globstar-new-bash-globbing-option. If false, `**` is treated like `*`. Default: true.
            case_insensitive: Whether the glob matching should be case insensitive. Default: false.
        """
        return self._bridge.run(
            self._async.expand_glob(
                glob,
                root=root,
                exclude=exclude,
                include_dirs=include_dirs,
                follow_symlinks=follow_symlinks,
                canonicalize=canonicalize,
                extended=extended,
                globstar=globstar,
                case_insensitive=case_insensitive,
            )
        )

    def link(self, target: str, path: str) -> None:
        """Create a hard link pointing to an existing file."""
        self._bridge.run(self._async.link(target, path))

    def lstat(self, path: str) -> FileInfo:
        """Return file information about a file or directory symlink."""
        return self._bridge.run(self._async.lstat(path))

    def make_temp_dir(
        self,
        *,
        dir: Optional[str] = None,
        prefix: Optional[str] = None,
        suffix: Optional[str] = None,
    ) -> str:
        """Create a new temporary directory.

        Args:
            dir: The directory where the temporary directory should be created.
            prefix: The prefix for the temporary directory name.
            suffix: The suffix for the temporary directory name.
        """
        return self._bridge.run(
            self._async.make_temp_dir(dir=dir, prefix=prefix, suffix=suffix)
        )

    def make_temp_file(
        self,
        *,
        dir: Optional[str] = None,
        prefix: Optional[str] = None,
        suffix: Optional[str] = None,
    ) -> str:
        """Create a new temporary file.

        Args:
            dir: The directory where the temporary file should be created.
            prefix: The prefix for the temporary file name.
            suffix: The suffix for the temporary file name.
        """
        return self._bridge.run(
            self._async.make_temp_file(dir=dir, prefix=prefix, suffix=suffix)
        )

    def read_link(self, path: str) -> str:
        """Read the target of a symbolic link."""
        return self._bridge.run(self._async.read_link(path))

    def real_path(self, path: str) -> str:
        """Return the canonicalized absolute pathname."""
        return self._bridge.run(self._async.real_path(path))

    def symlink(
        self,
        target: str,
        path: str,
        *,
        type: Optional[Literal["file", "dir", "junction"]] = None,
    ) -> None:
        """Create a symbolic link.

        Args:
            target: The target path of the symbolic link.
            path: The path where the symbolic link should be created.
            type: The type of the symbolic link (file, dir, or junction).
        """
        self._bridge.run(self._async.symlink(target, path, type=type))

    def truncate(self, name: str, length: Optional[int] = None) -> None:
        """Truncate or extend the specified file to reach a given size."""
        self._bridge.run(self._async.truncate(name, length))

    def umask(self, mask: Optional[int] = None) -> int:
        """Sets the process's file mode creation mask."""
        return self._bridge.run(self._async.umask(mask))

    def utime(self, path: str, atime: str, mtime: str) -> None:
        """Change the access and modification times of a file."""
        self._bridge.run(self._async.utime(path, atime, mtime))

    def upload(self, local_path: str, sandbox_path: str) -> None:
        """Upload a file, directory, or symlink from local filesystem to the sandbox."""
        self._bridge.run(self._async.upload(local_path, sandbox_path))

    def download(self, local_path: str, sandbox_path: str) -> None:
        """Download a file or directory from the sandbox."""
        self._bridge.run(self._async.download(local_path, sandbox_path))

    def create(self, path: str) -> FsFile:
        """Create a new, empty file at the specified path."""
        async_file = self._bridge.run(self._async.create(path))
        return FsFile(self._rpc, self._bridge, async_file._fd)

    def open(
        self,
        path: str,
        *,
        read: Optional[bool] = None,
        write: Optional[bool] = None,
        append: Optional[bool] = None,
        truncate: Optional[bool] = None,
        create: Optional[bool] = None,
        create_new: Optional[bool] = None,
        mode: Optional[int] = None,
    ) -> FsFile:
        """Open a file and return a file descriptor.

        Args:
            path: The path to the file to open.
            read: Sets the option for read access. This option, when `true`, means that the file should be read-able if opened. Default: true.
            write: Sets the option for write access. This option, when `true`, means that the file should be write-able if opened. Default: false.
            append: Sets the option for append mode. This option, when `true`, means that writes to the file will always append to the end. Default: false.
            truncate: If `true`, and the file already exists and is a regular file, it will be truncated to length 0 when opened. Default: false.
            create: If `true`, the file will be created if it does not already exist. Default: false.
            create_new: If `true`, the file will be created if it does not already exist. Default: false.
            mode: The permission mode to use when creating the file.
        """
        async_file = self._bridge.run(
            self._async.open(
                path,
                read=read,
                write=write,
                append=append,
                truncate=truncate,
                create=create,
                create_new=create_new,
                mode=mode,
            )
        )
        return FsFile(self._rpc, self._bridge, async_file._fd)



from __future__ import annotations

import base64
from typing import TYPE_CHECKING, AsyncIterable, BinaryIO, Iterable, Union

if TYPE_CHECKING:
    from .rpc import AsyncRpcClient

Streamable = Union[AsyncIterable[bytes], Iterable[bytes], BinaryIO]


class AsyncStreamWriter:
    """Manages writing a stream to the server."""

    def __init__(self, rpc: AsyncRpcClient, stream_id: int):
        self._rpc = rpc
        self._stream_id = stream_id

    async def start(self) -> None:
        """Send $sandbox.stream.start message."""
        await self._rpc.send_notification(
            "$sandbox.stream.start", {"streamId": self._stream_id}
        )

    async def enqueue(self, data: bytes) -> None:
        """Send $sandbox.stream.enqueue message with base64-encoded data."""
        await self._rpc.send_notification(
            "$sandbox.stream.enqueue",
            {
                "streamId": self._stream_id,
                "data": base64.b64encode(data).decode("ascii"),
            },
        )

    async def end(self) -> None:
        """Send $sandbox.stream.end message."""
        await self._rpc.send_notification(
            "$sandbox.stream.end", {"streamId": self._stream_id}
        )

    async def error(self, message: str) -> None:
        """Send $sandbox.stream.error message."""
        await self._rpc.send_notification(
            "$sandbox.stream.error", {"streamId": self._stream_id, "error": message}
        )


async def stream_data(
    rpc: AsyncRpcClient, data: Streamable, chunk_size: int = 64 * 1024
) -> int:
    """
    Stream data to server. Returns stream_id.

    Sends: start → enqueue(s) → end
    On error: sends error message then re-raises
    """
    stream_id, writer = await start_stream(rpc)

    try:
        await complete_stream(writer, data, chunk_size)
        return stream_id

    except Exception as e:
        await writer.error(str(e))
        raise


async def start_stream(rpc: AsyncRpcClient) -> tuple[int, AsyncStreamWriter]:
    """
    Start a stream and return (stream_id, writer).

    Only sends the start notification. Call complete_stream() to send data and end.
    """
    stream_id = rpc.next_stream_id()
    writer = AsyncStreamWriter(rpc, stream_id)
    await writer.start()
    return stream_id, writer


async def complete_stream(
    writer: AsyncStreamWriter, data: Streamable, chunk_size: int = 64 * 1024
) -> None:
    """
    Complete a stream by sending all data and the end notification.

    Should be called after start_stream() and after any RPC calls that need the stream_id.
    """
    if hasattr(data, "read"):
        # File-like object
        while True:
            chunk = data.read(chunk_size)  # type: ignore[union-attr]
            if not chunk:
                break
            await writer.enqueue(chunk)
    elif hasattr(data, "__aiter__"):
        # Async iterable
        async for chunk in data:  # type: ignore[union-attr]
            await writer.enqueue(chunk)
    else:
        # Sync iterable
        for chunk in data:
            await writer.enqueue(chunk)

    await writer.end()


def is_streamable(obj: object) -> bool:
    """Check if object is streamable (not bytes)."""
    return (
        hasattr(obj, "read")
        or hasattr(obj, "__aiter__")
        or (hasattr(obj, "__iter__") and not isinstance(obj, (bytes, str, dict, list)))
    )



from __future__ import annotations

from websockets import ClientConnection, ConnectionClosed, connect
from httpx import URL

from .errors import AuthenticationError


class WebSocketTransport:
    def __init__(self) -> None:
        self._ws: ClientConnection | None = None
        self._closed = False

    @property
    def closed(self) -> bool:
        return self._closed

    async def connect(self, url: URL, headers: dict[str, str]) -> ClientConnection:
        try:
            ws = await connect(str(url), additional_headers=headers)
            self._ws = ws
            return ws
        except Exception as e:
            if "HTTP 401" in str(e):
                raise AuthenticationError(
                    "Authentication failed, invalid API token"
                ) from e

            raise e

    async def send(self, data: str) -> None:
        if self._ws is None:
            raise RuntimeError("WebSocket is not connected")
        await self._ws.send(data)

    async def close(self) -> None:
        self._closed = True

        if self._ws is not None:
            await self._ws.close()

    async def __aexit__(self):
        await self.close()

    async def __aiter__(self):
        if self._ws is None:
            raise RuntimeError("WebSocket is not connected")

        try:
            async for message in self._ws:
                yield message
        except ConnectionClosed:
            return



from typing import TypedDict


class AuthenticationError(Exception):
    """Raised when there is an authentication error."""

    pass


class MissingApiToken(Exception):
    """Raised when an API token is missing."""

    pass


class UnknownRpcMethod(Exception):
    """Raised when an unknown RPC method is called."""

    pass


class ProcessAlreadyExited(Exception):
    """Raised when trying to interact with a process that has already exited."""

    pass


class HTTPStatusError(Exception):
    """Raised when an HTTP request returns a non-success status code."""

    def __init__(self, status_code: int, message: str) -> None:
        self.status_code = status_code
        self.message = message
        super().__init__(f"HTTP Status {status_code}: {message}")


class ZodErrorRaw(TypedDict):
    expected: str
    code: str
    path: list[str]
    message: str


class RpcValidationError(ValueError):
    """Raised when there is a validation error in RPC parameters."""

    def __init__(self, errors: list[ZodErrorRaw]) -> None:
        self.errors = errors

        msg = "Invalid Parameters:\n"
        for err in errors:
            msg += f"- Path: {'.'.join(err['path'])}, Expected: {err['expected']}\n"

        super().__init__(msg)



import re


def to_camel_case(snake_str):
    components = snake_str.split("_")
    return components[0] + "".join(x.title() for x in components[1:])


def convert_to_camel_case(data):
    if isinstance(data, dict):
        return {to_camel_case(k): convert_to_camel_case(v) for k, v in data.items()}
    elif isinstance(data, list):
        return [convert_to_camel_case(i) for i in data]
    else:
        return data


def to_snake_case(camel_str):
    return re.sub(r"(?<!^)(?=[A-Z])", "_", camel_str).lower()


def convert_to_snake_case(data):
    if isinstance(data, dict):
        return {to_snake_case(k): convert_to_snake_case(v) for k, v in data.items()}
    elif isinstance(data, list):
        return [convert_to_snake_case(i) for i in data]
    else:
        return data


def parse_link_header(header: str) -> dict[str, str]:
    links = {}
    parts = header.split(",")
    for part in parts:
        section = part.split(";")
        if len(section) < 2:
            continue
        url = section[0].strip()[1:-1]
        name = section[1].strip().split("=")[1][1:-1]
        links[name] = url
    return links



import asyncio
import threading
from typing import Any, Coroutine, TypeVar

T = TypeVar("T")


class AsyncBridge:
    def __init__(self):
        self.loop = asyncio.new_event_loop()
        self.thread = threading.Thread(target=self._run_loop, daemon=True)
        self.thread.start()

    def _run_loop(self) -> None:
        asyncio.set_event_loop(self.loop)
        try:
            self.loop.run_forever()
        finally:
            self.loop.close()

    def run(self, coro: Coroutine[Any, Any, T]) -> T:
        if not asyncio.iscoroutine(coro):
            raise TypeError("Must pass a coroutine")

        # Submit the coroutine to the loop in the other thread
        future = asyncio.run_coroutine_threadsafe(coro, self.loop)
        return future.result()

    def stop(self) -> None:
        self.loop.call_soon_threadsafe(self.loop.stop)
        self.thread.join()



from __future__ import annotations

from typing import Any, TypedDict
from typing_extensions import Optional


from .bridge import AsyncBridge
from .console import (
    AsyncConsoleClient,
    AsyncPaginatedList,
    PaginatedList,
)


class TimelineApp(TypedDict):
    id: str
    """The unique identifier for the app."""

    slug: str
    """The human readable identifier for the app."""


class TimelineContext(TypedDict):
    slug: str


class Domain(TypedDict):
    domain: str
    """The domain name."""


class Timeline(TypedDict):
    slug: str
    """The unique identifier for the timeline."""

    partition: dict[str, str]
    """The partition of the timeline."""

    app: TimelineApp
    context: TimelineContext
    """The context of the timeline."""

    domains: list[Domain]
    """The domains associated with the timeline."""


class AsyncTimelines:
    def __init__(self, client: AsyncConsoleClient):
        self._client = client

    async def list(
        self,
        app: str,
        *,
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> AsyncPaginatedList[Timeline]:
        """List timelines for a specific app.

        Args:
            app: The app ID or slug.
            cursor: The cursor for pagination.
            limit: Limit the number of items to return.
        """
        options: dict[str, Any] = {}
        if cursor is not None:
            options["cursor"] = cursor
        if limit is not None:
            options["limit"] = limit
        return await self._client.get_paginated(
            f"/api/v2/apps/{app}/timelines",
            cursor=None,
            params=options if options else None,
        )


class Timelines:
    def __init__(self, client: AsyncConsoleClient, bridge: AsyncBridge):
        self._client = client
        self._bridge = bridge
        self._async = AsyncTimelines(client)

    def list(
        self,
        app: str,
        *,
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> PaginatedList[Timeline]:
        """List timelines for a specific app.

        Args:
            app: The app ID or slug.
            cursor: The cursor for pagination.
            limit: Limit the number of items to return.
        """
        paginated = self._bridge.run(self._async.list(app, cursor=cursor, limit=limit))
        return PaginatedList(self._bridge, paginated)



from __future__ import annotations

import asyncio
import sys
from typing import Any, BinaryIO, Callable, Optional, TypedDict, TypeVar, cast
from typing_extensions import Literal

from .bridge import AsyncBridge
from .errors import ProcessAlreadyExited
from .rpc import (
    AsyncFetchResponse,
    AsyncRpcClient,
    FetchResponse,
)

T = TypeVar("T")


class AbortSignal:
    pass


class RemoteProcessOptions(TypedDict):
    stdout_inherit: bool
    stderr_inherit: bool


class ProcessSpawnResult(TypedDict):
    pid: int
    stdout_stream_id: int
    stderr_stream_id: int


class ProcessWaitResult(TypedDict):
    success: bool
    code: int
    signal: AbortSignal | None


class ChildProcessStatus(TypedDict):
    """Whether the subprocess exited successfully, i.e. with a zero exit code."""

    success: bool
    """The exit code of the subprocess. Negative values indicate that the process was killed by a signal."""
    code: int
    """The signal that caused the process to exit, if any. If the process exited normally, this will be `null`."""
    signal: AbortSignal | None


class AsyncChildProcess:
    def __init__(
        self,
        pid: int,
        stdout: asyncio.StreamReader,
        stderr: asyncio.StreamReader,
        wait_task: asyncio.Task,
        rpc: AsyncRpcClient,
        _stdout_task: Optional[asyncio.Task] = None,
        _stderr_task: Optional[asyncio.Task] = None,
        _process_list: Optional[list["AsyncChildProcess"]] = None,
    ):
        self.pid = pid
        self.stdout = stdout
        self.stderr = stderr
        self._wait_task = wait_task
        self.returncode = None
        self._rpc = rpc
        self._stdout_task = _stdout_task
        self._stderr_task = _stderr_task
        self._process_list = _process_list

    def _remove_from_list(self) -> None:
        """Remove this process from the tracking list."""
        if self._process_list is not None and self in self._process_list:
            self._process_list.remove(self)

    @classmethod
    async def create(
        cls: type["AsyncChildProcess"],
        res: ProcessSpawnResult,
        rpc: AsyncRpcClient,
        options: RemoteProcessOptions,
        process_list: Optional[list[AsyncChildProcess]] = None,
    ) -> AsyncChildProcess:
        return create_process_like(cls, res, rpc, options, process_list)

    async def wait(self) -> ChildProcessStatus:
        raw = await self._wait_task
        result = cast(ProcessWaitResult, raw)
        self._remove_from_list()
        return ChildProcessStatus(
            success=result["success"], code=result["code"], signal=result["signal"]
        )

    async def kill(self) -> None:
        """Kill the process and cancel all associated tasks."""

        try:
            await self._rpc.call("processKill", {"pid": self.pid})
        except ProcessAlreadyExited:
            # Process already exited
            pass

        # Cancel all associated tasks
        self._wait_task.cancel()
        if self._stdout_task is not None:
            self._stdout_task.cancel()
        if self._stderr_task is not None:
            self._stderr_task.cancel()

        self._remove_from_list()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.kill()


async def _pipe_stream(reader: asyncio.StreamReader, writer: BinaryIO):
    """Helper to pump data from the StreamReader to a local binary stream."""
    try:
        while not reader.at_eof():
            data = await reader.read(1024)
            if not data:
                break

            writer.write(data)
            writer.flush()
    except Exception:
        # Handle potential connection drops or closed pipes silently
        pass


def create_process_like(
    cls: Callable[
        [
            int,
            asyncio.StreamReader,
            asyncio.StreamReader,
            asyncio.Task,
            AsyncRpcClient,
            Optional[asyncio.Task],
            Optional[asyncio.Task],
            Optional[list[AsyncChildProcess]],
        ],
        T,
    ],
    res: ProcessSpawnResult,
    rpc: AsyncRpcClient,
    options: RemoteProcessOptions,
    process_list: Optional[list[AsyncChildProcess]] = None,
) -> T:
    pid = res["pid"]

    stdout = asyncio.StreamReader()
    stderr = asyncio.StreamReader()

    rpc._pending_processes[res["stdout_stream_id"]] = stdout
    rpc._pending_processes[res["stderr_stream_id"]] = stderr

    wait_task = rpc._loop.create_task(rpc.call("processWait", {"pid": pid}))

    stdout_task: Optional[asyncio.Task] = None
    stderr_task: Optional[asyncio.Task] = None

    if options.get("stdout_inherit"):
        stdout_coro = _pipe_stream(stdout, sys.stdout.buffer)
        stdout_task = rpc._loop.create_task(stdout_coro)
    if options.get("stderr_inherit"):
        stderr_coro = _pipe_stream(stderr, sys.stderr.buffer)
        stderr_task = rpc._loop.create_task(stderr_coro)

    instance = cls(
        pid, stdout, stderr, wait_task, rpc, stdout_task, stderr_task, process_list
    )

    return instance


class SyncStreamReader:
    """Wraps asyncio.StreamReader to provide synchronous read methods."""

    def __init__(self, bridge: AsyncBridge, reader: asyncio.StreamReader):
        self._bridge = bridge
        self._reader = reader

    def read(self, n: int = -1) -> bytes:
        return self._bridge.run(self._reader.read(n))

    def readline(self) -> bytes:
        return self._bridge.run(self._reader.readline())

    def readexactly(self, n: int) -> bytes:
        return self._bridge.run(self._reader.readexactly(n))


class ChildProcess:
    def __init__(
        self,
        rpc: AsyncRpcClient,
        bridge: AsyncBridge,
        async_proc: AsyncChildProcess,
    ):
        self._rpc = rpc
        self._bridge = bridge

        self._async_proc = async_proc
        self.stdout = SyncStreamReader(bridge, self._async_proc.stdout)
        self.stderr = SyncStreamReader(bridge, self._async_proc.stderr)
        self.returncode: int | None = None

    @property
    def pid(self) -> int:
        return self._async_proc.pid

    def wait(self) -> ChildProcessStatus:
        return self._bridge.run(self._async_proc.wait())

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._bridge.run(self._async_proc.__aexit__(exc_type, exc_val, exc_tb))


class AsyncDenoProcess(AsyncChildProcess):
    def __init__(
        self,
        pid: int,
        stdout: asyncio.StreamReader,
        stderr: asyncio.StreamReader,
        wait_task: asyncio.Task,
        rpc: AsyncRpcClient,
        stdout_task: Optional[asyncio.Task],
        stderr_task: Optional[asyncio.Task],
        process_list: Optional[list[AsyncChildProcess]] = None,
    ):
        super().__init__(
            pid, stdout, stderr, wait_task, rpc, stdout_task, stderr_task, process_list
        )
        self._listening_task: asyncio.Task | None = None

    @classmethod
    async def create(
        cls: type["AsyncDenoProcess"],
        res: ProcessSpawnResult,
        rpc: AsyncRpcClient,
        options: RemoteProcessOptions,
        process_list: Optional[list[AsyncChildProcess]] = None,
    ) -> AsyncDenoProcess:
        p = create_process_like(cls, res, rpc, options, process_list)

        p._listening_task = rpc._loop.create_task(
            rpc.call("denoHttpWait", {"pid": p.pid})
        )
        return p

    async def kill(self) -> None:
        """Kill the process and cancel all associated tasks."""
        if self._listening_task is not None:
            self._listening_task.cancel()
        await super().kill()

    async def wait_http_ready(self) -> bool:
        """Whether the Deno process is ready to accept HTTP requests."""

        if self._listening_task is not None:
            return await self._listening_task

        return False

    async def fetch(
        self,
        url: str,
        method: Optional[str] = "GET",
        headers: Optional[dict[str, str]] = None,
        redirect: Optional[Literal["follow", "manual"]] = None,
    ) -> AsyncFetchResponse:
        """Fetch a URL from the Deno process."""
        return await self._rpc.fetch(url, method, headers, redirect, self.pid)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self._listening_task is not None:
            self._listening_task.cancel()
        await super().__aexit__(exc_type, exc_val, exc_tb)


class DenoProcess(ChildProcess):
    def __init__(
        self,
        rpc: AsyncRpcClient,
        bridge: AsyncBridge,
        async_proc: AsyncDenoProcess,
    ):
        super().__init__(rpc, bridge, async_proc)
        self._async = async_proc

    def wait_http_ready(self) -> bool:
        """Whether the Deno process is ready to accept HTTP requests."""
        return self._bridge.run(self._async.wait_http_ready())

    def fetch(
        self,
        url: str,
        method: Optional[str] = "GET",
        headers: Optional[dict[str, str]] = None,
        redirect: Optional[Literal["follow", "manual"]] = None,
    ) -> FetchResponse:
        """Fetch a URL from the Deno process."""
        async_response = self._bridge.run(
            self._rpc.fetch(url, method, headers, redirect, self.pid)
        )
        return FetchResponse(async_response)


class AsyncDenoRepl(AsyncChildProcess):
    @classmethod
    async def create(
        cls: type["AsyncDenoRepl"],
        res: ProcessSpawnResult,
        rpc: AsyncRpcClient,
        options: RemoteProcessOptions,
        process_list: Optional[list[AsyncChildProcess]] = None,
    ) -> AsyncDenoRepl:
        return create_process_like(cls, res, rpc, options, process_list)

    async def eval(self, code: str) -> str:
        """Evaluate code in the REPL and return the output."""

        result = await self._rpc.call("denoReplEval", {"code": code, "pid": self.pid})

        return result

    async def call(self, fn: str, args: list[Any]) -> Any:
        """Call a function in the REPL process."""

        result = await self._rpc.call(
            "denoReplCall", {"pid": self.pid, "fn": fn, "args": args}
        )

        return result

    async def close(self) -> None:
        """Close the REPL process."""

        await self._rpc.call("denoReplClose", {"pid": self.pid})
        self._wait_task.cancel()
        if self._stdout_task is not None:
            self._stdout_task.cancel()
        if self._stderr_task is not None:
            self._stderr_task.cancel()
        self._remove_from_list()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()


class DenoRepl:
    def __init__(
        self,
        rpc: AsyncRpcClient,
        bridge: AsyncBridge,
        async_proc: AsyncDenoRepl,
    ):
        self._rpc = rpc
        self._bridge = bridge

        self._async = async_proc
        self.stdout = SyncStreamReader(bridge, self._async.stdout)
        self.stderr = SyncStreamReader(bridge, self._async.stderr)

    def eval(self, code: str) -> str:
        """Evaluate code in the REPL and return the output."""
        return self._bridge.run(self._async.eval(code))

    def call(self, fn: str, args: list[Any]) -> Any:
        """Call a function in the REPL process."""
        return self._bridge.run(self._async.call(fn, args))

    def close(self) -> None:
        """Close the REPL process."""
        self._bridge.run(self._async.close())

    @property
    def returncode(self) -> int | None:
        return self._async.returncode

    @property
    def pid(self) -> int:
        return self._async.pid

    def wait(self) -> ChildProcessStatus:
        return self._bridge.run(self._async.wait())

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._bridge.run(self._async.__aexit__(exc_type, exc_val, exc_tb))



from __future__ import annotations

from typing import Any, TypedDict, cast
from typing_extensions import Optional

from .utils import convert_to_snake_case
from .bridge import AsyncBridge
from .console import (
    PaginatedList,
    AsyncConsoleClient,
    AsyncPaginatedList,
)


class App(TypedDict):
    id: str
    """The unique identifier for the app."""

    slug: str
    """The human readable identifier for the app."""

    created_at: str
    """The ISO 8601 timestamp when the app was created."""

    updated_at: str
    """The ISO 8601 timestamp when the app was last updated."""


class AsyncApps:
    def __init__(self, client: AsyncConsoleClient):
        self._client = client

    async def get(self, id_or_slug: str) -> App | None:
        """Get an app by its ID or slug."""
        result = await self._client.get_or_none(f"/api/v2/apps/{id_or_slug}")
        if result is None:
            return None
        raw_result = convert_to_snake_case(result)
        return cast(App, raw_result)

    async def list(
        self,
        *,
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> AsyncPaginatedList[App]:
        """List apps of an org.

        Args:
            cursor: The cursor for pagination.
            limit: Limit the number of items to return.
        """
        options: dict[str, Any] = {}
        if cursor is not None:
            options["cursor"] = cursor
        if limit is not None:
            options["limit"] = limit
        return await self._client.get_paginated(
            "/api/v2/apps", cursor=None, params=options if options else None
        )

    async def create(
        self,
        *,
        slug: Optional[str] = None,
    ) -> App:
        """Create a new app.

        Args:
            slug: Human readable identifier for the app.
        """
        options: dict[str, Any] = {}
        if slug is not None:
            options["slug"] = slug
        result = await self._client.post("/api/v2/apps", options)
        raw_result = convert_to_snake_case(result)
        return cast(App, raw_result)

    async def update(
        self,
        app: str,
        *,
        slug: Optional[str] = None,
    ) -> App:
        """Update an existing app.

        Args:
            app: The app ID or slug to update.
            slug: Human readable identifier for the app.
        """
        update: dict[str, Any] = {}
        if slug is not None:
            update["slug"] = slug
        result = await self._client.patch(f"/api/v2/apps/{app}", update)
        raw_result = convert_to_snake_case(result)
        return cast(App, raw_result)

    async def delete(self, app: str) -> None:
        """Delete an app by its ID or slug."""
        await self._client.delete(f"/api/v2/apps/{app}")


class Apps:
    def __init__(self, client: AsyncConsoleClient, bridge: AsyncBridge):
        self._client = client
        self._bridge = bridge
        self._async = AsyncApps(client)

    def get(self, id_or_slug: str) -> App | None:
        """Get an app by its ID or slug."""
        return self._bridge.run(self._async.get(id_or_slug))

    def list(
        self,
        *,
        cursor: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> PaginatedList[App]:
        """List apps of an org.

        Args:
            cursor: The cursor for pagination.
            limit: Limit the number of items to return.
        """
        paginated = self._bridge.run(self._async.list(cursor=cursor, limit=limit))
        return PaginatedList(self._bridge, paginated)

    def create(
        self,
        *,
        slug: Optional[str] = None,
    ) -> App:
        """Create a new app.

        Args:
            slug: Human readable identifier for the app.
        """
        return self._bridge.run(self._async.create(slug=slug))

    def update(
        self,
        app: str,
        *,
        slug: Optional[str] = None,
    ) -> App:
        """Update an existing app.

        Args:
            app: The app ID or slug to update.
            slug: Human readable identifier for the app.
        """
        return self._bridge.run(self._async.update(app, slug=slug))

    def delete(self, app: str) -> None:
        """Delete an app by its ID or slug."""
        self._bridge.run(self._async.delete(app))
